from gcloudc.db.models.fields.charfields import CharField
from unittest import skip
from django.db.models import Model

import sleuth
from django.db import connection
from django.db.utils import IntegrityError

from gcloudc.db.backends.datastore.transaction import TransactionFailedError

from . import TestCase
from .models import (
    TestUser,
    TestUserTwo,
)


def _get_client():
    return connection.connection.gclient


def get_kind_query(kind, keys_only=True):
    datastore_client = _get_client()
    query = datastore_client.query(kind=kind)
    if keys_only:
        query.keys_only()
    return list(query.fetch())


class TestUniqueConstraints(TestCase):

    KINDS_TO_DELETE = ["uniquemarker", "test_testuser", "test_testusertwo"]

    def test_insert(self):
        """
        Assert that when creating a new instance, unique markers are also
        created to reflect the constraints defined on the model.

        If a subsequent insert is attempted, these should be compared to
        enforce a constraint similar to SQL.
        """
        TestUser.objects.create(username="tommyd", first_name="Tommy", second_name="Shelby")

        # attempt to create another entity which violates one of the constraints
        with self.assertRaises(IntegrityError):
            TestUser.objects.create(username="tommyd", first_name="Tommy", second_name="Doherty")

    def test_insert_unique_together(self):
        """
        Assert that when creating a new instance, unique markers are also
        created to reflect the constraints defined on the model.

        If a subsequent insert is attempted, these should be compared to
        enforce a constraint similar to SQL.
        """
        TestUser.objects.create(username="tommyd", first_name="Tommy", second_name="Doherty")

        # attempt to create another entity which violates a unique_together constraint
        with self.assertRaises(IntegrityError):
            TestUser.objects.create(username="thetommyd", first_name="Tommy", second_name="Doherty")

    def test_bulk_insert(self):
        """
        Assert that bulk inserts respect any unique markers made inside
        the same transaction.
        """
        with self.assertRaises(IntegrityError):
            TestUserTwo.objects.bulk_create(
                [
                    TestUserTwo(username="Mickey Bell"),
                    TestUserTwo(username="Tony Thorpe"),
                    TestUserTwo(username="Mickey Bell"),
                ]
            )

        self.assertEqual(TestUserTwo.objects.count(), 0)

        # sanity check normal bulk insert works
        TestUserTwo.objects.bulk_create([TestUserTwo(username="Mickey Bell"), TestUserTwo(username="Tony Thorpe")])
        self.assertEqual(TestUserTwo.objects.count(), 2)

        # and if we were to run the bulk insert, previously created
        # unique markers are still respected
        with self.assertRaises(IntegrityError):
            TestUserTwo.objects.bulk_create([TestUserTwo(username="Mickey Bell"), TestUserTwo(username="Tony Thorpe")])
        self.assertEqual(TestUserTwo.objects.count(), 2)

    def test_update_with_constraint_conflict(self):
        TestUserTwo.objects.create(username="AshtonGateEight")
        user_two = TestUserTwo.objects.create(username="AshtonGateSeven")

        # now do the update operation
        user_two.username = "AshtonGateEight"
        with self.assertRaises(IntegrityError):
            user_two.save()

    def test_update_with_constraint_together_conflict(self):
        TestUser.objects.create(username="tommyd", first_name="Tommy", second_name="Doherty")
        user_two = TestUser.objects.create(username="tommye", first_name="Tommy", second_name="Einfield")

        # now do the update operation
        user_two.second_name = "Doherty"
        with self.assertRaises(IntegrityError):
            user_two.save()

    def test_error_on_update_does_not_change_entity(self):
        """
        Assert that when there is an error / exception raised as part of the
        update command, the entity is rolled back to its originial state.
        """
        user = TestUserTwo.objects.create(username="AshtonGateEight")

        with sleuth.detonate("gcloudc.db.backends.datastore.transaction.Transaction.put", TransactionFailedError):
            with self.assertRaises(TransactionFailedError):
                user.username = "Red Army"
                user.save()

        user.refresh_from_db()
        self.assertEqual(user.username, "AshtonGateEight")

    def test_bulk_update(self):
        """
        Assert that updates via the QuerySet API handle uniques.
        """
        user_one = TestUser.objects.create(username="stevep", first_name="steve", second_name="phillips")
        user_two = TestUser.objects.create(username="joeb", first_name="joe", second_name="burnell")

        # now do the update operation on the queryset
        TestUser.objects.all().update(first_name="lee")

        user_one.refresh_from_db()
        user_two.refresh_from_db()

        self.assertEqual(user_one.first_name, "lee")
        self.assertEqual(user_two.first_name, "lee")

    def test_error_with_bulk_update(self):
        user_one = TestUser.objects.create(username="stevep", first_name="steve", second_name="phillips")
        user_two = TestUser.objects.create(username="joeb", first_name="joe", second_name="burnell")

        with self.assertRaises(IntegrityError):
            TestUser.objects.all().update(username="stevep")

        user_one.refresh_from_db()
        user_two.refresh_from_db()

        # in Djangae (python 2) this doesn't work, user_two would end up
        # with username=bill, which makes it non transactional on the group
        self.assertEqual(user_one.username, "stevep")
        self.assertEqual(user_two.username, "joeb")

    def test_error_with_bulk_update_in_memory(self):
        user_one = TestUser.objects.create(username="stevep", first_name="steve", second_name="phillips")
        user_two = TestUser.objects.create(username="joeb", first_name="joe", second_name="burnell")

        with self.assertRaises(IntegrityError):
            TestUser.objects.all().update(username="bill")

        user_one.refresh_from_db()
        user_two.refresh_from_db()

        self.assertEqual(user_one.username, "stevep")
        self.assertEqual(user_two.username, "joeb")

    def test_error_with_bulk_update_unique_together(self):
        user_one = TestUser.objects.create(username="stevep", first_name="steve", second_name="phillips")
        user_two = TestUser.objects.create(username="joeb", first_name="joe", second_name="burnell")

        with self.assertRaises(IntegrityError):
            TestUser.objects.all().update(first_name="lee", second_name="bruce")

        user_one.refresh_from_db()
        user_two.refresh_from_db()

        # in djangae (python 2) this doesn't work, user_two would end up
        # with username=bill, which makes it non transactional on the group
        self.assertEqual(user_one.first_name, "steve")
        self.assertEqual(user_two.first_name, "joe")

    def test_error_with_bulk_update_unique_together_in_memory(self):
        user_one = TestUser.objects.create(username="stevem", first_name="steve", second_name="mitchell")
        user_two = TestUser.objects.create(username="joem", first_name="joe", second_name="mitchell")

        with self.assertRaises(IntegrityError):
            TestUser.objects.all().update(first_name="lee")

        user_one.refresh_from_db()
        user_two.refresh_from_db()

        self.assertEqual(user_one.first_name, "steve")
        self.assertEqual(user_two.first_name, "joe")

    # see https://github.com/googleapis/google-cloud-python/issues/9921
    @skip("This test should (probably) not fail once the emulator bug is fixed")
    def test_500_limit(self):
        # TODO: datastore emulator seems to fail at the old 25 limit, update
        # this test once emulator issue is addressed
        for i in range(25):
            username = "stevep_{}".format(i)
            first_name = "steve_{}".format(i)
            second_name = "phillips_{}".format(i)
            TestUser.objects.create(
                username=username,
                first_name=first_name,
                second_name=second_name,
            )

        TestUser.objects.all().update(first_name="lee")

        for i in range(25, 501):
            username = "stevep_{}".format(i)
            first_name = "steve_{}".format(i)
            second_name = "phillips_{}".format(i)
            TestUser.objects.create(
                username=username,
                first_name=first_name,
                second_name=second_name,
            )

        # This should raise because of the 500 changes per transaction limit
        with self.assertRaises(IntegrityError):
            TestUser.objects.all().update(first_name="lee")

    def test_bulk_delete_fails_if_limit_exceeded(self):
        """
        Assert that there is currently a practical limitation when deleting multi
        entities, based on a combination of the unique markers per model
        and transaction limit of touching 500 entities.
        """
        TestUserTwo.objects.create(username="Mickey Bell")
        TestUserTwo.objects.create(username="Tony Thorpe")

        with sleuth.switch("gcloudc.db.backends.datastore.transaction.TRANSACTION_ENTITY_LIMIT", 1):
            with self.assertRaises(Exception):
                TestUserTwo.objects.all().delete()

    def test_delete_entity_fails(self):
        """
        Assert that if the entity delete operation fails, the user is not deleted.
        """
        user = TestUserTwo.objects.create(username="Mickey Bell")

        with sleuth.detonate(
            "gcloudc.db.backends.datastore.commands.remove_entities_from_cache_by_key", TransactionFailedError
        ):
            with self.assertRaises(TransactionFailedError):
                user.delete()

        # the entity in question should not have been deleted, as error in the
        # transactions atomic block should revert all changes
        user.refresh_from_db()

    def test_polymodels_with_base_unique(self):
        """
        Test that a polymodel unique constraint doesn't blow when the parent
        class has a unique constraint
        """
        class Base(Model):
            unique_field = CharField(unique=True)

        class Child(Base):
            pass

        # This used to raise an integrity error because the field was checked
        # for both base and Child class
        child = Child.objects.create(unique_field="unique_value")
        self.assertIsNotNone(child)

        # Check that actual integrity issues are reported
        with self.assertRaises(IntegrityError):
            Child.objects.create(unique_field="unique_value")

    def test_polymodels_with_child_unique(self):
        """
        Test that a polymodel unique constraint doesn't blow when the child
        class has a unique constraint
        """
        class Base(Model):
            pass

        class Child(Base):
            unique_field = CharField(unique=True)

        child = Child.objects.create(unique_field="unique_value")
        self.assertIsNotNone(child)

        with self.assertRaises(IntegrityError):
            Child.objects.create(unique_field="unique_value")

    def test_polymodels_with_base_unique_together(self):
        """
        Test that a polymodel unique_together constraint is respected when
        in the parent model
        """
        class Base(Model):
            a = CharField()
            b = CharField()

            class Meta():
                unique_together = ["a", "b"]

        class Child(Base):
            pass

        child = Child.objects.create(a="a", b="b")
        self.assertIsNotNone(child)

        # Check that actual integrity issues are reported
        with self.assertRaises(IntegrityError):
            child = Child.objects.create(a="a", b="b")

    def test_polymodels_with_child_unique_together(self):
        """
        Test that a polymodel unique_together constraint is respected when
        in the child model
        """
        class Base(Model):
            pass

        class Child(Base):
            a = CharField()
            b = CharField()

            class Meta():
                unique_together = ["a", "b"]

        child = Child.objects.create(a="a", b="b")
        self.assertIsNotNone(child)

        # Check that actual integrity issues are reported
        with self.assertRaises(IntegrityError):
            child = Child.objects.create(a="a", b="b")

    def test_polymodels_with_cross_model_unique_together(self):
        """
        Test that a polymodel unique_together constraint is respected when
        it references fields across the hierarchy

        Note: this wouldn't work with a SQL backend because unique constraint
        are per-table. Since we are doing read-before-write to enforce  unique
        constraint, table structures is not a problem.
        """
        class Base(Model):
            a = CharField()

        class Child(Base):
            b = CharField()

            class Meta():
                unique_together = ["a", "b"]

        child = Child.objects.create(a="a", b="b")
        self.assertIsNotNone(child)

        # Check that actual integrity issues are reported
        with self.assertRaises(IntegrityError):
            child = Child.objects.create(a="a", b="b")

    def test_unique_in_abstract_parent(self):
        """
        Test that a polymodel unique constraint doesn't blow when the parent
        class has a unique constraint
        """
        class Base(Model):
            unique_field = CharField(unique=True)

            class Meta():
                abstract = True

        class Child(Base):
            pass

        child = Child.objects.create(unique_field="unique_value")
        self.assertIsNotNone(child)

        # Check that actual integrity issues are reported
        with self.assertRaises(IntegrityError):
            Child.objects.create(unique_field="unique_value")




import functools

try:
    from collections.abc import Hashable  # Python >=3.10
except ImportError:
    from collections import Hashable


class memoized(object):
    def __init__(self, func, *args):
        self.func = func
        self.cache = {}
        self.args = args

    def __call__(self, *args):
        args = self.args or args
        if not isinstance(args, Hashable):
            # uncacheable. a list, for instance.
            # better to not cache than blow up.
            return self.func(*args)

        if args in self.cache:
            return self.cache[args]
        else:
            value = self.func(*args)
            self.cache[args] = value
            return value

    def __repr__(self):
        """Return the function's docstring."""
        return self.func.__doc__

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return functools.partial(self.__call__, obj)



import logging
import threading

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from . import utils
from .context import (
    ContextCache,
    key_or_entity_compare,
)
from .unique_utils import (
    _format_value_for_identifier,
    unique_identifiers_from_entity,
)

_local = threading.local()

logger = logging.getLogger("djangae")


def get_context():
    try:
        return _local.context
    except AttributeError:
        _local.context = ContextCache()
        return _local.context


CACHE_ENABLED = getattr(settings, "GCLOUDC_CACHE_ENABLED", True)

# The max number of entities in a resultset that will be cached.
# If a query returns more than this number then only the first ones
# will be cached
DEFAULT_MAX_ENTITY_COUNT = 8
MAX_CACHE_COUNT = getattr(settings, "GCLOUDC_CACHE_MAX_ENTITY_COUNT", DEFAULT_MAX_ENTITY_COUNT)


class CachingSituation:
    DATASTORE_GET = 0
    DATASTORE_PUT = 1
    DATASTORE_GET_PUT = 2  # When we are doing an update


def _apply_namespace(value_or_map, namespace):
    """ Add the given namespace to the given cache key(s). """
    if hasattr(value_or_map, "keys"):
        return {"{}:{}".format(namespace, k): v for k, v in value_or_map.items()}
    elif hasattr(value_or_map, "__iter__") and not isinstance(value_or_map, str):
        return ["{}:{}".format(namespace, x) for x in value_or_map]
    else:
        return "{}:{}".format(namespace, value_or_map)


def _strip_namespace(value_or_map):
    """ Remove the namespace part from the given cache key(s). """

    def _strip(value):
        return value.split(":", 1)[-1]

    if hasattr(value_or_map, "keys"):
        return {_strip(k): v for k, v in value_or_map.items()}
    elif hasattr(value_or_map, "__iter__"):
        return [_strip(x) for x in value_or_map]
    else:
        return _strip(value_or_map)


def _get_cache_key_and_model_from_datastore_key(key):
    from django.db.migrations.recorder import MigrationRecorder

    # The django migration model isn't registered with the app registry so this
    # is special cased here
    MODELS_WHICH_ARENT_REGISTERED_WITH_DJANGO = {
        MigrationRecorder.Migration._meta.db_table: MigrationRecorder.Migration
    }

    kind = key.kind()
    model = utils.get_model_from_db_table(kind)

    if not model:
        if kind in MODELS_WHICH_ARENT_REGISTERED_WITH_DJANGO:
            model = MODELS_WHICH_ARENT_REGISTERED_WITH_DJANGO[kind]
        else:
            # This should never happen.. if it does then we can edit get_model_from_db_table to pass
            # include_deferred=True/included_swapped=True to get_models, whichever makes it better
            raise ImproperlyConfigured(
                "Unable to locate model for db_table '{}' - are you missing an INSTALLED_APP?".format(key.kind())
            )

    # We build the cache key for the ID of the instance
    cache_key = "|".join(
        [key.kind(), "{}:{}".format(model._meta.pk.column, _format_value_for_identifier(key.id_or_name()))]
    )

    return (cache_key, model)


def add_entities_to_cache(model, entities, situation, namespace):
    from gcloudc.db.transaction import in_atomic_block

    if not CACHE_ENABLED:
        return None

    context = get_context()

    if not context.context_enabled:
        # Don't cache anything if caching is disabled
        return

    # Don't cache on Get if we are inside a transaction, even in the context
    # This is because transactions don't see the current state of the datastore
    # We can still cache in the context on Put()
    if situation == CachingSituation.DATASTORE_GET and in_atomic_block():
        return

    identifiers = [unique_identifiers_from_entity(model, entity) for entity in entities]

    for ent_identifiers, entity in zip(identifiers, entities):
        get_context().stack.top.cache_entity(_apply_namespace(ent_identifiers, namespace), entity, situation)


def remove_entities_from_cache_by_key(keys, namespace):
    """
        Given an iterable of datastore.Keys objects, remove the corresponding entities from cache
    """

    if not CACHE_ENABLED:
        return None

    context = get_context()

    for key in keys:
        identifiers = context.stack.top.cache.get_reversed(key, compare_func=key_or_entity_compare)

        for identifier in identifiers:
            if identifier in context.stack.top.cache:
                del context.stack.top.cache[identifier]


def get_from_cache_by_key(key):
    """
        Given a datastore.Key (which should already have the namespace applied to it), return an
        entity from the context cache
    """

    if not CACHE_ENABLED:
        return None

    context = get_context()
    ret = None
    if context.context_enabled:
        # It's safe to hit the context cache, because a new one was pushed on the stack at the start of the transaction
        ret = context.stack.top.get_entity_by_key(key)

    return ret


def get_from_cache(unique_identifier, namespace):
    """
        Return an entity from the context cache
    """
    context = get_context()

    if not CACHE_ENABLED:
        return None

    cache_key = _apply_namespace(unique_identifier, namespace)
    ret = None
    if context.context_enabled:
        # It's safe to hit the context cache, because a new one was pushed on the stack at the start of the transaction
        ret = context.stack.top.get_entity(cache_key)

    return ret


def reset_context(keep_disabled_flags=False, *args, **kwargs):
    """
        Called at the beginning and end of each request, resets the thread local
        context. If you pass keep_disabled_flags=True the context_enabled
        flags will be preserved, this is really only useful for testing.
    """

    context = get_context()
    context.reset(keep_disabled_flags=keep_disabled_flags)



import pickle
from datetime import timedelta

from django import forms
from django.core.exceptions import ValidationError
from django.db import (
    connection,
    models,
)
from django.test import override_settings
from gcloudc.db.models.fields.charfields import (
    CharField,
    CharOrNoneField,
)
from gcloudc.db.models.fields.computed import (
    ComputedBooleanField,
    ComputedCharField,
    ComputedIntegerField,
    ComputedPositiveIntegerField,
    ComputedTextField,
)

from gcloudc.db.models.fields.iterable import SetField, ListField
from gcloudc.db.models.fields.related import RelatedSetField, RelatedListField, GenericRelationField
from gcloudc.db.models.fields.json import JSONField

from . import TestCase
from .models import (
    BasicTestModel,
    BinaryFieldModel,
    ModelWithCharField,
    NonIndexedModel,
    PFAwards,
    PFAuthor,
    PFPost,
    ISOther,
    ISStringReferenceModel,
)


class BasicTest(TestCase):
    def test_basic_connector_usage(self):
        # Create
        instance = BasicTestModel.objects.create(field1="Hello World!", field2=1998)

        # Count
        self.assertEqual(1, BasicTestModel.objects.count())

        # Get
        self.assertEqual(instance, BasicTestModel.objects.get())

        # Update
        instance.field1 = "Hello Mars!"
        instance.save()

        # Query
        instance2 = BasicTestModel.objects.filter(field1="Hello Mars!")[0]

        self.assertEqual(instance, instance2)
        self.assertEqual(instance.field1, instance2.field1)

        # Query by PK
        instance2 = BasicTestModel.objects.filter(pk=instance.pk)[0]

        self.assertEqual(instance, instance2)
        self.assertEqual(instance.field1, instance2.field1)

        # Non-existent PK
        instance3 = BasicTestModel.objects.filter(pk=999).first()
        self.assertIsNone(instance3)

        # Unique field
        instance2 = BasicTestModel.objects.filter(field2=1998)[0]

        self.assertEqual(instance, instance2)
        self.assertEqual(instance.field1, instance2.field1)


class CharFieldModelTests(TestCase):
    def test_char_field_with_max_length_set(self):
        test_bytestrings = [(u"01234567891", 11), (u"ążźsęćńół", 17)]

        for test_text, byte_len in test_bytestrings:
            test_instance = ModelWithCharField(char_field_with_max=test_text)
            self.assertRaisesMessage(
                ValidationError,
                "Ensure this value has at most 10 bytes (it has %d)." % byte_len,
                test_instance.full_clean,
            )

    def test_char_field_with_not_max_length_set(self):
        longest_valid_value = u"0123456789" * 150
        too_long_value = longest_valid_value + u"more"

        test_instance = ModelWithCharField(char_field_without_max=longest_valid_value)
        test_instance.full_clean()  # max not reached so it's all good

        test_instance.char_field_without_max = too_long_value
        self.assertRaisesMessage(
            ValidationError, u"Ensure this value has at most 1500 bytes (it has 1504).", test_instance.full_clean
        )

    def test_additional_validators_work(self):
        test_instance = ModelWithCharField(char_field_as_email="bananas")
        self.assertRaisesMessage(ValidationError, "failed", test_instance.full_clean)

    def test_too_long_max_value_set(self):
        try:

            class TestModel(models.Model):
                test_char_field = CharField(max_length=1501)

        except AssertionError as e:
            self.assertEqual(str(e), "CharFields max_length must not be greater than 1500 bytes.")


class ModelWithCharOrNoneField(models.Model):
    char_or_none_field = CharOrNoneField(max_length=100)


class CharOrNoneFieldTests(TestCase):
    def test_char_or_none_field(self):
        # Ensure that empty strings are coerced to None on save
        obj = ModelWithCharOrNoneField.objects.create(char_or_none_field="")
        obj.refresh_from_db()
        self.assertIsNone(obj.char_or_none_field)


class StringReferenceRelatedSetFieldModelTests(TestCase):
    def test_can_update_related_field_from_form(self):
        related = ISOther.objects.create()
        thing = ISStringReferenceModel.objects.create(related_things={related})
        before_set = thing.related_things
        thing.related_list.field.save_form_data(thing, set())
        thing.save()
        self.assertNotEqual(before_set.all(), thing.related_things.all())

    def test_saving_forms(self):
        class TestForm(forms.ModelForm):
            class Meta:
                model = ISStringReferenceModel
                fields = ("related_things",)

        related = ISOther.objects.create()
        post_data = {"related_things": [str(related.pk)]}

        form = TestForm(post_data)
        self.assertTrue(form.is_valid())
        instance = form.save()
        self.assertEqual({related.pk}, instance.related_things_ids)


class RelatedFieldPrefetchTests(TestCase):
    def test_prefetch_related(self):
        award = PFAwards.objects.create(name="award")
        author = PFAuthor.objects.create(awards={award})
        PFPost.objects.create(authors={author})

        posts = list(PFPost.objects.all().prefetch_related("authors__awards"))

        with self.assertNumQueries(0):
            list(posts[0].authors.all()[0].awards.all())


class PickleTests(TestCase):
    def test_all_fields_are_pickleable(self):
        """ In order to work with Djangae's migrations, all fields must be pickeable. """
        fields = [
            CharField(),
            CharOrNoneField(),
            ComputedBooleanField("method_name"),
            ComputedCharField("method_name"),
            ComputedIntegerField("method_name"),
            ComputedPositiveIntegerField("method_name"),
            ComputedTextField("method_name"),
            GenericRelationField(),
            JSONField(default=list),
            ListField(CharField(), default=["badger"]),
            SetField(CharField(), default=set(["badger"])),
        ]

        fields.extend(
            [RelatedListField(ModelWithCharField), RelatedSetField(ModelWithCharField)]
        )

        for field in fields:
            try:
                pickle.dumps(field)
            except (pickle.PicklingError, TypeError) as e:
                self.fail("Could not pickle %r: %s" % (field, e))


class BinaryFieldModelTests(TestCase):
    binary_value = b"\xff"

    def test_insert(self):

        obj = BinaryFieldModel.objects.create(binary=self.binary_value)
        obj.save()

        readout = BinaryFieldModel.objects.get(pk=obj.pk)

        assert readout.binary == self.binary_value

    def test_none(self):

        obj = BinaryFieldModel.objects.create()
        obj.save()

        readout = BinaryFieldModel.objects.get(pk=obj.pk)

        assert readout.binary is None

    def test_update(self):

        obj = BinaryFieldModel.objects.create()
        obj.save()

        toupdate = BinaryFieldModel.objects.get(pk=obj.pk)
        toupdate.binary = self.binary_value
        toupdate.save()

        readout = BinaryFieldModel.objects.get(pk=obj.pk)

        assert readout.binary == self.binary_value


class CharFieldModel(models.Model):
    char_field = models.CharField(max_length=500)


class CharFieldModelTest(TestCase):
    def test_query(self):
        instance = CharFieldModel(char_field="foo")
        instance.save()

        readout = CharFieldModel.objects.get(char_field="foo")
        self.assertEqual(readout, instance)

    def test_query_unicode(self):
        name = u"Jacqu\xe9s"

        instance = CharFieldModel(char_field=name)
        instance.save()

        readout = CharFieldModel.objects.get(char_field=name)
        self.assertEqual(readout, instance)

    @override_settings(DEBUG=True)
    def test_query_unicode_debug(self):
        """ Test that unicode query can be performed in DEBUG mode,
            which will use CursorDebugWrapper and call last_executed_query.
        """
        name = u"Jacqu\xe9s"

        instance = CharFieldModel(char_field=name)
        instance.save()

        readout = CharFieldModel.objects.get(char_field=name)
        self.assertEqual(readout, instance)


class DurationFieldModelWithDefault(models.Model):
    duration = models.DurationField(default=timedelta(1, 0))


class DurationFieldModelTests(TestCase):
    def test_creates_with_default(self):
        instance = DurationFieldModelWithDefault()

        self.assertEqual(instance.duration, timedelta(1, 0))

        instance.save()

        readout = DurationFieldModelWithDefault.objects.get(pk=instance.pk)
        self.assertEqual(readout.duration, timedelta(1, 0))

    def test_none_saves_as_default(self):
        instance = DurationFieldModelWithDefault()
        # this could happen if we were reading an existing instance out of the database that didn't have this field
        instance.duration = None
        instance.save()

        readout = DurationFieldModelWithDefault.objects.get(pk=instance.pk)
        self.assertEqual(readout.duration, timedelta(1, 0))


class ModelWithNonNullableFieldAndDefaultValue(models.Model):
    some_field = models.IntegerField(null=False, default=1086)


class NonIndexedModelFieldsTests(TestCase):
    def test_long_textfield(self):
        long_text = "A" * 1501
        instance = NonIndexedModel()
        instance.content = long_text
        instance.save()

    def test_big_binaryfield(self):
        long_binary = ("A" * 1501).encode('utf-8')
        instance = NonIndexedModel()
        instance.binary = long_binary
        instance.save()


# ModelWithNonNullableFieldAndDefaultValueTests verifies that we maintain same
# behavior as Django with respect to a model field that is non-nullable with default value.
class ModelWithNonNullableFieldAndDefaultValueTests(TestCase):
    def _create_instance_with_null_field_value(self):

        instance = ModelWithNonNullableFieldAndDefaultValue.objects.create(some_field=1)
        client = connection.connection.gclient
        entity = client.get(
            client.key(
                ModelWithNonNullableFieldAndDefaultValue._meta.db_table,
                instance.pk,
                namespace=connection.settings_dict.get("NAMESPACE", ""),
            )
        )
        del entity["some_field"]
        client.put(entity)

        instance.refresh_from_db()

        return instance

    def test_none_in_db_reads_as_none_in_model(self):
        instance = self._create_instance_with_null_field_value()
        self.assertIsNone(instance.some_field)

    def test_none_in_model_saved_as_default(self):

        instance = self._create_instance_with_null_field_value()

        instance.save()
        instance.refresh_from_db()

        self.assertEqual(instance.some_field, 1086)



from __future__ import unicode_literals

import json

SELECT_PATTERN = """
SELECT (%(columns)s) FROM %(table)s
WHERE %(where)s
ORDER BY %(order)s
OFFSET %(offset)s
LIMIT %(limit)s
""".lstrip()

INSERT_PATTERN = """
INSERT INTO %(table)s (%(columns)s)
VALUES %(values)s
""".lstrip()

UPDATE_PATTERN = """
REPLACE INTO %(table)s (%(columns)s)
VALUES %(values)s
WHERE %(where)s
""".lstrip()

DELETE_PATTERN = """
DELETE FROM %(table)s
WHERE %(where)s
""".lstrip()


def _generate_values_expression(instances, columns):
    values = []
    for instance in instances:
        row = []
        for column in columns:
            # FIXME: should this use get_default as a default?
            value = getattr(instance, column, None)

            try:
                text_value = _quote_string(value)
            except UnicodeDecodeError:
                text_value = "'<binary>'"

            row.append(text_value)

        values.append("(" + ", ".join(row) + ")")
    return ", ".join(values)


def _generate_insert_sql(command):
    columns = sorted([x.column for x in command.fields])

    params = {
        "table": command.model._meta.db_table,
        "columns": ", ".join(columns),
        "values": _generate_values_expression(command.objs, columns),
    }

    return (INSERT_PATTERN % params).replace("\n", " ").strip()


def _generate_where_expression(representation):
    where = []
    for branch in representation["where"]:
        branch = "(" + " AND ".join(["%s%s" % (k, _quote_string(v)) for k, v in branch.items()]) + ")"
        where.append(branch)

    return " OR ".join(where)


def _quote_string(value):
    needs_quoting = isinstance(value, str)
    # in ANSI SQL as well as GQL, string literals are wrapped in single quotes
    return "'{}'".format(value) if needs_quoting else str(value)


def _generate_select_sql(command, representation):
    has_offset = representation["low_mark"] > 0
    has_limit = representation["high_mark"] is not None
    has_ordering = bool(representation["order_by"])
    has_where = bool(representation["where"])

    lines = SELECT_PATTERN.split("\n")

    # Remove limit and offset and where if we don't need them
    if not has_limit:
        del lines[4]

    if not has_offset:
        del lines[3]

    if not has_ordering:
        del lines[2]

    if not has_where:
        del lines[1]

    sql = "\n".join(lines)

    columns = (
        "*"
        if not representation["columns"]
        else ", ".join(sorted(representation["columns"]))  # Just to make the output predictable
    )

    ordering = [
        ("%s %s" % (x.lstrip("-"), "DESC" if x.startswith("-") else "")).strip() for x in representation["order_by"]
    ]

    replacements = {
        "table": representation["table"],
        "columns": columns,
        "offset": representation["low_mark"],
        "limit": (representation["high_mark"] or 0) - (representation["low_mark"] or 0),
        "where": _generate_where_expression(representation),
        "order": ", ".join(ordering),
    }

    return (sql % replacements).replace("\n", " ").strip()


def _generate_delete_sql(command, representation):
    has_where = bool(representation["where"])

    lines = DELETE_PATTERN.split("\n")
    if not has_where:
        del lines[1]

    sql = "\n".join(lines)

    return (
        (sql % {"table": representation["table"], "where": _generate_where_expression(representation)})
        .replace("\n", " ")
        .strip()
    )


def _generate_update_sql(command, representation):
    has_where = bool(representation["where"])

    lines = UPDATE_PATTERN.split("\n")
    if not has_where:
        del lines[2]

    sql = "\n".join(lines)
    columns = sorted([x[0].column for x in command.values])

    values = {x[0].column: str(x[2]) for x in command.values}

    params = {
        "table": representation["table"],
        "columns": ", ".join(columns),
        "values": "(" + ", ".join([values[x] for x in columns]) + ")",
        "where": _generate_where_expression(representation),
    }

    return (sql % params).replace("\n", " ").strip()


def generate_sql_representation(command):
    from .commands import SelectCommand, DeleteCommand, UpdateCommand, InsertCommand

    if isinstance(command, InsertCommand):
        # Inserts don't have a .query so we have to deal with them
        # separately
        return _generate_insert_sql(command)

    representation = json.loads(command.query.serialize())

    if isinstance(command, SelectCommand):
        return _generate_select_sql(command, representation)
    elif isinstance(command, DeleteCommand):
        return _generate_delete_sql(command, representation)
    elif isinstance(command, UpdateCommand):
        return _generate_update_sql(command, representation)

    raise NotImplementedError("Unrecognized query type")



import json
import logging
import os
import subprocess
import time
from datetime import datetime
from urllib.error import (
    HTTPError,
    URLError,
)
from urllib.request import urlopen

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.core.management import load_command_class
from django.utils.autoreload import DJANGO_AUTORELOAD_ENV

_COMPONENTS_LIST_COMMAND = "gcloud components list --format=json".split()
_REQUIRED_COMPONENTS = set(["beta", "cloud-datastore-emulator", "core"])

_BASE_COMMAND = "gcloud beta emulators datastore start --user-output-enabled=false --consistency=1.0 --quiet --project=test".split()  # noqa
_DEFAULT_PORT = 9090

logger = logging.getLogger(__name__)


class CloudDatastoreRunner:
    USE_MEMORY_DATASTORE_BY_DEFAULT = False

    def __init__(self, *args, **kwargs):
        self._process = None
        super().__init__(*args, **kwargs)

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument("--no-datastore", action="store_false", dest="datastore", default=True)
        parser.add_argument("--datastore-port", action="store", dest="port", default=_DEFAULT_PORT)
        parser.add_argument(
            "--use-memory-datastore",
            action="store_true",
            dest="use_memory_datastore",
            default=self.USE_MEMORY_DATASTORE_BY_DEFAULT,
        )

    def execute(self, *args, **kwargs):
        try:
            if kwargs.get("datastore", True) and os.environ.get(DJANGO_AUTORELOAD_ENV) != "true":
                self._check_gcloud_components()
                self._start_emulator(**kwargs)

            super().execute(*args, **kwargs)
        finally:
            self._stop_emulator()

    def _check_gcloud_components(self):
        finished_process = subprocess.run(_COMPONENTS_LIST_COMMAND, stdout=subprocess.PIPE, encoding="utf-8")
        installed_components = set(
            [cp["id"] for cp in json.loads(finished_process.stdout) if cp["current_version_string"] is not None]
        )

        if not _REQUIRED_COMPONENTS.issubset(installed_components):
            raise RuntimeError(
                "Missing Google Cloud SDK component(s): {}\n"
                "Please run `gcloud components install` to install missing components.".format(
                    ", ".join(_REQUIRED_COMPONENTS - installed_components)
                )
            )

    def _datastore_filename(self):
        BASE_DIR = getattr(settings, "BASE_DIR", None)

        if not BASE_DIR:
            raise ImproperlyConfigured("Please define BASE_DIR in your Django settings")

        return os.path.join(BASE_DIR, ".datastore")

    def _get_args(self, **kwargs):
        args = ["--host-port=127.0.0.1:%s" % kwargs.get("port", _DEFAULT_PORT)]

        if kwargs["use_memory_datastore"]:
            logger.info("Using in-memory datastore")
            args.append("--no-store-on-disk")
        else:
            args.append("--data-dir=%s" % self._datastore_filename())

        return args

    def _wait_for_datastore(self, **kwargs):
        TIMEOUT = 60.0

        start = datetime.now()

        logger.info("Waiting for Cloud Datastore Emulator...")
        time.sleep(1)

        failures = 0
        while True:
            try:
                response = urlopen("http://127.0.0.1:%s/" % kwargs["port"])
            except (HTTPError, URLError):
                failures += 1
                time.sleep(1)
                if failures > 5:
                    # Only start logging if this becomes persistent
                    logger.exception(
                        "Error connecting to the Cloud Datastore Emulator. Retrying...")
                continue

            if response.status == 200:
                # Give things a second to really boot
                time.sleep(2)
                break

            if (datetime.now() - start).total_seconds() > TIMEOUT:
                raise RuntimeError("Unable to start Cloud Datastore Emulator. Please check the logs.")

            time.sleep(1)

    def _start_emulator(self, **kwargs):
        logger.info("Starting Cloud Datastore Emulator")

        os.environ["DATASTORE_EMULATOR_HOST"] = "127.0.0.1:%s" % kwargs["port"]
        os.environ["DATASTORE_PROJECT_ID"] = "test"

        # The Cloud Datastore emulator regularly runs out of heap space
        # so set a higher max
        os.environ["JDK_JAVA_OPTIONS"] = "-Xms512M -Xmx1024M"

        env = os.environ.copy()
        self._process = subprocess.Popen(_BASE_COMMAND + self._get_args(**kwargs), env=env)

        self._wait_for_datastore(**kwargs)

    def _stop_emulator(self):
        logger.info("Stopping Cloud Datastore Emulator")
        if self._process:
            self._process.terminate()
            self._process.wait()
            self._process = None


def locate_command(name):
    """
        Apps may override Django commands, what we want to do is
        subclass whichever one had precedence before the gcloudc.commands app and subclass that
    """

    try:
        index = settings.INSTALLED_APPS.index("gcloudc.commands")
    except ValueError:
        raise ImproperlyConfigured("Unable to locate gcloudc.commands in INSTALLED_APPS")

    APPS_TO_CHECK = list(settings.INSTALLED_APPS) + ["django.core"]

    for i in range(index + 1, len(APPS_TO_CHECK)):
        app_label = APPS_TO_CHECK[i]
        try:
            command = load_command_class(app_label, name)
        except ModuleNotFoundError:
            continue

        if command:
            return command.__class__
    else:
        raise ImportError("Unable to locate a base %s Command to subclass" % name)



import threading
import functools


# Because decorators are only instantiated once per function, we need to make sure any state
# stored on them is both thread-local (to prevent function calls in different threads
# interacting with each other) and safe to use recursively (by using a stack of state)


class ContextState(object):
    "Stores state per-call of the ContextDecorator"
    pass


class ContextDecorator(object):
    """
        A thread-safe ContextDecorator. Subclasses should implement classmethods
        called _do_enter(state, decorator_args) and _do_exit(state, decorator_args, exception)

        state is a thread.local which can store state for each enter/exit. Decorator args holds
        any arguments passed into the decorator or context manager when called.
    """

    VALID_ARGUMENTS = tuple()

    def __init__(self, func=None, **kwargs):
        # Func will be passed in if this has been called without parenthesis
        # as a @decorator

        # Make sure only valid decorator arguments were passed in
        if len(kwargs) > len(self.__class__.VALID_ARGUMENTS):
            raise ValueError(
                "Unexpected decorator arguments: {}".format(set(kwargs.keys()) - set(self.__class__.VALID_ARGUMENTS))
            )

        self.func = func
        self.decorator_args = {x: kwargs.get(x) for x in self.__class__.VALID_ARGUMENTS}
        # Add thread local state for variables that change per-call rather than
        # per insantiation of the decorator
        self.state = threading.local()
        self.state.stack = []

    def __get__(self, obj, objtype=None):
        """ Implement descriptor protocol to support instance methods. """
        # Invoked whenever this is accessed as an attribute of *another* object
        # - as it is when wrapping an instance method: `instance.method` will be
        # the ContextDecorator, so this is called.
        # We make sure __call__ is passed the `instance`, which it will pass onto
        # `self.func()`
        return functools.partial(self.__call__, obj)

    def __call__(self, *args, **kwargs):
        # Called if this has been used as a decorator not as a context manager

        def decorated(*_args, **_kwargs):
            decorator_args = self.decorator_args.copy()
            exception = False
            self.__class__._do_enter(self._push_state(), decorator_args)
            try:
                return self.func(*_args, **_kwargs)
            except Exception:
                exception = True
                raise
            finally:
                self.__class__._do_exit(self._pop_state(), decorator_args, exception)

        if not self.func:
            # We were instantiated with args
            self.func = args[0]
            return decorated
        else:
            return decorated(*args, **kwargs)

    def _push_state(self):
        "We need a stack for state in case a decorator is called recursively"
        # self.state is a threading.local() object, so if the current thread is not the one in
        # which ContextDecorator.__init__ was called (e.g. is not the thread in which the function
        # was decorated), then the 'stack' attribute may not exist
        if not hasattr(self.state, "stack"):
            self.state.stack = []

        self.state.stack.append(ContextState())
        return self.state.stack[-1]

    def _pop_state(self):
        return self.state.stack.pop()

    def __enter__(self):
        return self.__class__._do_enter(self._push_state(), self.decorator_args.copy())

    def __exit__(self, exc_type, exc_value, traceback):
        self.__class__._do_exit(self._pop_state(), self.decorator_args.copy(), exc_type)



import json
from . import TestCase
from unittest.mock import patch
from gcloudc.commands.management.commands import _REQUIRED_COMPONENTS, CloudDatastoreRunner


class CloudDatastoreRunnerTest(TestCase):
    def test_check_gcloud_components(self):
        class MockProcess:
            stdout = json.dumps([{"id": cp, "current_version_string": "0.1"} for cp in list(_REQUIRED_COMPONENTS)[:-1]])

        # We mock _start_emulator as we don't want to get that far in execute()
        with patch(
            "gcloudc.commands.management.commands.CloudDatastoreRunner._start_emulator",
            side_effect=AssertionError("Google Cloud components check failed"),
        ):

            with patch("gcloudc.commands.management.commands.subprocess.run", return_value=MockProcess()):
                command = CloudDatastoreRunner()
                with self.assertRaises(RuntimeError):
                    command.execute()



from hashlib import md5


def _has_enabled_constraints(model_or_instance):
    """
    Since we're no longer doing any extra write to enforce uniqueness there's
    not much point in not enforcing it, so we always keep it enabled
    """
    return True


def _has_unique_constraints(model_or_instance):
    """
    Returns a boolean to indicate if the given model has any type of unique
    constraint defined (e.g. unique on a single field, or meta.unique_together).

    To support concrete model inheritance we state that uniqueness checks
    should only be performed on the class that the defines the unique constraint.

    Note - you can see a much more verbose implementation of this in
    django.db.models.base.Model._get_unique_checks() - but we implement our
    own logic to exit early when the first constraint is found.
    """
    meta_options = model_or_instance._meta
    # we may get an instance here, so ensure we have a reference to the
    # model class
    model_class = meta_options.model
    unique_together = meta_options.unique_together
    unique_fields = any(
        field.unique and field.model == model_class
        for field in meta_options.fields
    )

    return any([unique_fields, unique_together])


def query_is_unique(model, query):
    """
        If the query is entirely on unique constraints then return the unique identifier for
        that unique combination. Otherwise return False
    """

    from . import meta_queries  # Circular import workaround

    if isinstance(query, meta_queries.AsyncMultiQuery):
        # By definition, a multiquery is not unique
        return False

    combinations = _unique_combinations(model)

    query_by_keys = {"{} {}".format(x[0], x[1]): x[2] for x in query.filters}
    for combination in combinations:
        unique_match = True
        field_names = []
        for field in combination:
            if field == model._meta.pk.column:
                field = "__key__"
            else:
                field = model._meta.get_field(field).column
            field_names.append(field)

            # We don't match this combination if the field didn't exist in the queried fields
            # or if it was, but the value was None (you can have multiple NULL values, they aren't unique)
            matching_filters = list(filter(lambda x: x[0] == field and x[1] == '=', query.filters))
            if len(matching_filters) != 1 or matching_filters[0][2] is None:
                unique_match = False
                break

        if unique_match:
            return "|".join(
                [model._meta.db_table]
                + ["{}:{}".format(x, _format_value_for_identifier(
                    query_by_keys["{} =".format(x)])) for x in field_names]
            )

    return False


def unique_identifiers_from_entity(model, entity, ignore_pk=True, ignore_null_values=True):
    """
    This method returns a list of all unique marker key identifier values for
    the given entity by combining the field and entity values. For example:

    [
        # example of a single field unique constraint
        djange_<model_db_table>|<field_name>:<entity_value>
        # example of unique together
        djange_<model_db_table>|<field_name>:<entity_value>|<field_name>:<entity_value>
        ...
    ]

    These are then used before we put() anything into the database, to check
    that there are no existing markers satisfying those unique constraints.
    """
    from .utils import get_top_concrete_parent

    # get all combintatons of unique combinations defined on the model class
    unique_combinations = _unique_combinations(model, ignore_pk)

    meta = model._meta

    identifiers = []
    for combination in unique_combinations:
        combo_identifiers = [[]]

        include_combination = True

        for field_name in combination:
            field = meta.get_field(field_name)

            if field.primary_key:
                value = entity.key.id_or_name
            else:
                value = entity.get(field.column)  # Get the value from the entity

            # If ignore_null_values is True, then we don't include combinations where the value is None
            # or if the field is a multivalue field where None means no value (you can't store None in a list)
            if (value is None and ignore_null_values) or (not value and isinstance(value, (list, set))):
                include_combination = False
                break

            if not isinstance(value, (list, set)):
                value = [value]

            new_combo_identifers = []

            for existing in combo_identifiers:
                for v in value:
                    identifier = "{}:{}".format(field.column, _format_value_for_identifier(v))
                    new_combo_identifers.append(existing + [identifier])

            combo_identifiers = new_combo_identifers

        if include_combination:
            # create the final value - eg <app_db_table>|<field_name>:<field_val>
            for identifier_pairs in combo_identifiers:
                constraint_prefix = get_top_concrete_parent(model)._meta.db_table
                constraint_suffix = "|".join(identifier_pairs)
                constraint_value = "{prefix}|{suffix}".format(prefix=constraint_prefix, suffix=constraint_suffix)
                identifiers.append(constraint_value)

    return identifiers


def _unique_combinations(model, ignore_pk=False):
    """
    Returns an iterable of iterables to represent all the unique constraints
    defined on the model. For example given the following model definition:

        class ExampleModel(models.Model):
            username = models.CharField(unique=True)
            email = models.EmailField(primary_key=True)
            first_name = models.CharField()
            second_name = models.CharField()

        class Meta:
            unique_together = ('first_name', 'second_name')

    This method would return

    [
        ['username'], # from field level unique=True
        ['email'], # implicit unique constraint from primary_key=True
        ['first_name', 'second_name'] # from model meta unique_together
    ]

    Fields with unique constraint defined in a concrete parent model are ingored
    since they're checked when that model is saved
    """
    # first grab all the unique together constraints
    unique_constraints = [list(together_constraint) for together_constraint in model._meta.unique_together]

    # then the column level constraints - special casing PK if required
    for field in model._meta.fields:
        if field.primary_key and ignore_pk:
            continue

        if field.unique and field.model == model:
            unique_constraints.append([field.name])

    # the caller should sort each inner iterable really - but we do this here
    # for now - motive being that interpolated keys from these values are consistent
    return [sorted(constraint) for constraint in unique_constraints]


def _format_value_for_identifier(value):
    # AppEngine max key length is 500 chars, so if the value is a string we hexdigest it to reduce the length
    # otherwise we str() it as it's probably an int or bool or something.
    return md5(value.encode("utf-8")).hexdigest() if isinstance(value, str) else str(value)



import logging

from django.core.exceptions import FieldError, FieldDoesNotExist, EmptyResultSet
from django.db import NotSupportedError
from django.db.models.aggregates import Aggregate
from django.db.models.expressions import Star
from django.db.models.query import FlatValuesListIterable, QuerySet
from django.db.models.sql.query import Query as DjangoQuery

from ..query import Query, WhereNode
from ..utils import get_top_concrete_parent


# Django >= 1.9
class ValuesListQuerySet(object):
    pass


def _iterable(x):
    # Returns True if x is iterable and not a string type
    return hasattr(x, "__iter__") and not isinstance(x, (str, bytes))


logger = logging.getLogger(__name__)


INVALID_ORDERING_FIELD_MESSAGE = (
    "Ordering on TextField or BinaryField is not supported on the Datastore. "
    "You might consider using a ComputedCharField which stores the first "
    "_MAX_STRING_LENGTH (from google.appengine.api.datastore_types) bytes of the "
    "field and instead order on that."
)


def _get_concrete_fields_with_model(model):
    return [
        (f, f.model if f.model != model else None)
        for f in model._meta.get_fields()
        if f.concrete and (not f.is_relation or f.one_to_one or (f.many_to_one and f.related_model))
    ]


def _walk_django_where(query, trunk_callback, leaf_callback, **kwargs):
    """
        Walks through a Django where tree. If a leaf node is encountered
        the leaf_callback is called, otherwise the trunk_callback is called
    """

    def walk_node(node, **kwargs):
        negated = kwargs["negated"]

        if node.negated:
            negated = not negated

        for child in node.children:
            new_kwargs = kwargs.copy()
            new_kwargs["negated"] = negated
            if not getattr(child, "children", []):
                leaf_callback(child, **new_kwargs)
            else:
                new_parent = trunk_callback(child, **new_kwargs)

                if new_parent:
                    new_kwargs["new_parent"] = new_parent

                walk_node(child, **new_kwargs)

    kwargs.setdefault("negated", False)
    walk_node(query.where, **kwargs)


class BaseParser(object):
    """
        Supports the most recent version of Django. Subclasses can
        override parts of this to support older versions.

        (NOTE: this goes against all my thoughts on good OO but I can't think
        of a cleaner way of doing things at the moment!)
    """

    def __init__(self, django_query, connection=None):
        self.django_query = django_query
        self.connection = connection
        self.model = self.django_query.model

    def _determine_query_kind(self):
        """ Basically returns SELECT or COUNT """
        query = self.django_query

        if query.annotations:
            if "__count" in query.annotations:
                field = query.annotations["__count"].source_expressions[0]
                if isinstance(field, Star) or field.value == "*":
                    return "COUNT"

        return "SELECT"

    def _prepare_for_transformation(self):
        from django.db.models.sql.where import NothingNode

        query = self.django_query

        def where_will_always_be_empty(where):
            if isinstance(where, NothingNode):
                return True

            if where.connector == "AND" and any(isinstance(x, NothingNode) for x in where.children):
                return True

            if where.connector == "OR" and len(where.children) == 1 and isinstance(where.children[0], NothingNode):
                return True

            return False

        # It could either be a NothingNode, or a WhereNode(AND NothingNode)
        if where_will_always_be_empty(query.where):
            # Empty where means return nothing!
            raise EmptyResultSet()

    def _extract_projected_columns(self):
        result = []
        query = self.django_query

        if query.select:
            for x in query.select:
                column = x.target.column
                result.append(column)
            return result
        else:
            # If the query uses defer()/only() then we need to process deferred. We have to get all deferred columns
            # for all (concrete) inherited models and then only include columns if they appear in that list
            only_load = query.get_loaded_field_names()
            if only_load:
                for field, model in _get_concrete_fields_with_model(query.model):
                    model = model or query.model
                    try:
                        if field.column in only_load[model]:
                            # Add a field that has been explicitly included
                            result.append(field.column)
                    except KeyError:
                        # Model wasn't explicitly listed in the only_load table
                        # Therefore, we need to load all fields from this model
                        result.append(field.column)
                return result
            else:
                return []

    def _apply_ordering_to_query(self, query):
        # Extract the ordering of the query results
        for order_col in self.get_extracted_ordering():
            query.add_order_by(order_col)

    def _set_projected_columns_on_query(self, query):
        # Extract any projected columns (values/values_list/only/defer)
        for projected_col in self._extract_projected_columns():
            query.add_projected_column(projected_col)

    def _apply_extra_selects_to_query(self, query):
        # Add any extra selects
        for col, select in self.django_query.extra_select.items():
            query.add_extra_select(col, select[0])

    def _apply_distinct_columns_to_query(self, query):
        if self.django_query.distinct:
            # This must happen after extracting projected cols
            query.set_distinct(list(self.django_query.distinct_fields))

    def _apply_annotations_to_query(self, query):
        # Process annotations!
        if self.django_query.annotation_select:
            for k, v in self.django_query.annotation_select.items():
                query.add_annotation(k, v)

    def _where_node_trunk_callback(self, node, negated, new_parent, **kwargs):
        new_node = WhereNode(new_parent.using)
        new_node.connector = node.connector
        new_node.negated = node.negated

        new_parent.children.append(new_node)

        return new_node

    def _where_node_leaf_callback(self, node, negated, new_parent, connection, model, compiler):
        new_node = WhereNode(new_parent.using)

        def convert_rhs_op(node):
            db_rhs = getattr(node.rhs, "_db", None)
            if db_rhs is not None and db_rhs != connection.alias:
                raise ValueError(
                    "Subqueries aren't allowed across different databases. Force "
                    "the inner query to be evaluated using `list(inner_query)`."
                )

            value = node.get_rhs_op(connection, node.rhs)
            operator = value.split()[0].lower().strip()
            if operator == "between":
                operator = "range"
            return operator

        if not hasattr(node, "lhs"):
            raise NotSupportedError("Attempted probable subquery, these aren't supported on the Datastore")

        # Don't call on querysets
        if not hasattr(node.rhs, "_as_sql") and not isinstance(node.rhs, DjangoQuery):
            try:
                # Although we do nothing with this. We need to call it as many lookups
                # perform validation etc.
                node.process_rhs(compiler, connection)
            except EmptyResultSet:
                if node.lookup_name == "in":
                    node.rhs = []
                else:
                    raise

        # Leaf
        if hasattr(node.lhs, "target"):
            # from Django 1.9, some node.lhs might not have a target attribute
            # as they might be wrapping date fields
            field = node.lhs.target
            operator = convert_rhs_op(node)
        elif isinstance(node.lhs, Aggregate):
            raise NotSupportedError("Aggregate filters are not supported on the Datastore")
        else:
            field = node.lhs.lhs.target
            operator = convert_rhs_op(node)

            # This deals with things like datefield__month__gt=X which means from this point
            # on, operator will have two parts in that particular case and will probably need to
            # be dealt with by a special indexer
            if node.lookup_name != node.lhs.lookup_name:
                operator = "{}__{}".format(node.lhs.lookup_name, node.lookup_name)

        if get_top_concrete_parent(field.model) != get_top_concrete_parent(model):
            raise NotSupportedError("Cross-join where filters are not supported on the Datastore")

        # Make sure we don't let people try to filter on a text field, otherwise they just won't
        # get any results!

        lookup_supports_text = getattr(node, "lookup_supports_text", False)

        if field.db_type(connection) in ("bytes", "text") and not lookup_supports_text:
            raise NotSupportedError("You can't filter on text or blob fields on the Datastore")

        if operator == "isnull" and field.model._meta.parents.values():
            raise NotSupportedError("isnull lookups on inherited relations aren't supported on the Datastore")

        lhs = field.column

        if hasattr(node.rhs, "get_compiler"):
            if len(node.rhs.select) == 1:
                # In Django >= 1.11 this is a values list type query, which we explicitly handle
                # because of the common case of pk__in=Something.objects.values_list("pk", flat=True)
                qs = QuerySet(query=node.rhs, using=self.connection.alias)

                # We make the query for the values, but wrap in a list to trick the
                # was_iter code below. This whole set of if/elif statements needs rethinking!
                rhs = [list(qs.values_list("pk", flat=True))]
            else:
                # This is a subquery
                raise NotSupportedError("Attempted to run a subquery on the Datastore")
        elif isinstance(node.rhs, ValuesListQuerySet):
            # We explicitly handle ValuesListQuerySet because of the
            # common case of pk__in=Something.objects.values_list("pk", flat=True)
            # this WILL execute another query, but that is to be expected on a
            # non-relational database.

            rhs = [x for x in node.rhs]  # Evaluate the queryset

        elif isinstance(node.rhs, QuerySet):
            # In Django 1.9, ValuesListQuerySet doesn't exist anymore, and instead
            # values_list returns a QuerySet
            if node.rhs._iterable_class == FlatValuesListIterable:
                # if the queryset has FlatValuesListIterable as iterable class
                # then it's a flat list, and we just need to evaluate the
                # queryset converting it into a list
                rhs = [x for x in node.rhs]
            else:
                # otherwise, we try to get the PK from the queryset
                rhs = list(node.rhs.values_list("pk", flat=True))
        else:
            rhs = node.rhs

        was_iter = _iterable(node.rhs)
        rhs = node.get_db_prep_lookup(rhs, connection)[-1]
        if rhs and not was_iter and _iterable(rhs):
            rhs = rhs[0]

        new_node.set_leaf(
            lhs,
            operator,
            rhs,
            is_pk_field=field == model._meta.pk,
            negated=negated,
            lookup_name=node.lookup_name,
            namespace=connection.namespace,
            target_field=field,
        )

        # For some reason, this test:
        # test_update_with_related_manager (get_or_create.tests.UpdateOrCreateTests)
        # ends up with duplicate nodes in the where tree. I don't know why. But this
        # weirdly causes the Datastore query to return nothing.
        # so here we don't add duplicate nodes, I can't think of a case where that would
        # change the query if it's under the same parent.
        if new_node in new_parent.children:
            return

        new_parent.children.append(new_node)

    def _generate_where_node(self, query):
        output = WhereNode(query.connection.alias)
        output.connector = self.django_query.where.connector

        _walk_django_where(
            self.django_query,
            self._where_node_trunk_callback,
            self._where_node_leaf_callback,
            new_parent=output,
            connection=self.connection,
            negated=self.django_query.where.negated,
            model=self.model,
            compiler=self.django_query.get_compiler(self.connection.alias),
        )

        return output

    def get_transformed_query(self):
        self._prepare_for_transformation()

        kind = self._determine_query_kind()

        ret = Query(self.model, kind)
        ret.connection = self.connection

        # Add the root concrete table as the source table
        root_table = get_top_concrete_parent(self.model)._meta.db_table
        ret.add_source_table(root_table)

        self._apply_ordering_to_query(ret)
        self._set_projected_columns_on_query(ret)
        self._apply_extra_selects_to_query(ret)
        self._apply_distinct_columns_to_query(ret)
        self._apply_annotations_to_query(ret)

        # Extract any query offsets and limits
        ret.low_mark = self.django_query.low_mark
        ret.high_mark = self.django_query.high_mark

        output = self._generate_where_node(ret)

        # If there no child nodes, just wipe out the where
        if not output.children:
            output = None

        ret.where = output
        return ret

    def get_extracted_ordering(self):
        from ..commands import log_once
        from django.db.models.expressions import OrderBy, F
        from ..expressions import Scatter

        query = self.django_query

        # Add any orderings
        if not query.default_ordering:
            result = list(query.order_by)
        else:
            result = list(query.order_by or query.get_meta().ordering or [])

        if query.extra_order_by:
            result = list(query.extra_order_by)

            # we need some extra logic to handle dot seperated ordering
            new_result = []
            cross_table_ordering = set()
            for ordering in result:
                if "." in ordering:
                    dot_based_ordering = ordering.split(".")
                    if dot_based_ordering[0] == query.model._meta.db_table:
                        ordering = dot_based_ordering[1]
                    elif dot_based_ordering[0].lstrip("-") == query.model._meta.db_table:
                        ordering = "-{}".format(dot_based_ordering[1])
                    else:
                        cross_table_ordering.add(ordering)
                        continue  # we don't want to add this ordering value
                new_result.append(ordering)

            if len(cross_table_ordering):
                log_once(
                    logger.warning,
                    "The following orderings were ignored as cross-table orderings"
                    " are not supported on the Datastore: %s",
                    cross_table_ordering,
                )

            result = new_result

        final = []

        opts = query.model._meta

        # Apparently expression ordering is absolute and so shouldn't be flipped
        # if the standard_ordering is False. This keeps track of which columns
        # were expressions and so don't need flipping
        expressions = set()

        for col in result:
            if isinstance(col, OrderBy):
                descending = col.descending
                col = col.expression.name
                if descending:
                    col = "-" + col
                expressions.add(col)
            elif isinstance(col, F):
                col = col.name
            elif isinstance(col, Scatter):
                final.append("__scatter__")
                continue

            if isinstance(col, int):
                # If you do a Dates query, the ordering is set to [1] or [-1]... which is weird
                # I think it's to select the column number but then there is only 1 column so
                # unless the ordinal is one-based I have no idea. So basically if it's an integer
                # subtract 1 from the absolute value and look up in the select for the column (guessing)
                idx = abs(col) - 1
                try:
                    field_name = query.select[idx].col.col[-1]
                    field = query.model._meta.get_field(field_name)
                    final.append("-" + field.column if col < 0 else field.column)
                except IndexError:
                    raise NotSupportedError("Unsupported order_by %s" % col)
            elif col.lstrip("-") == "pk":
                pk_col = "__key__"
                final.append("-" + pk_col if col.startswith("-") else pk_col)
            elif col == "?":
                raise NotSupportedError("Random ordering is not supported on the Datastore")
            elif col.lstrip("-").startswith("__") and col.endswith("__"):
                # Allow stuff like __scatter__
                final.append(col)
            elif "__" in col:
                continue
            else:
                try:
                    column = col.lstrip("-")

                    # This is really 1.8 only, but I didn't want to duplicate this function
                    # just for this. Suggestions for doing this more cleanly welcome!
                    if column in getattr(query, "annotation_select", {}):
                        # It's an annotation, if it's a supported one, return the
                        # original column
                        annotation = query.annotation_select[column]
                        name = annotation.__class__.__name__

                        # We only support a few expressions
                        if name not in ("Trunc", "Col", "Date", "DateTime"):
                            raise NotSupportedError("Tried to order by unsupported expression")
                        elif name == "Trunc":
                            column = annotation.lhs.output_field.column
                        else:
                            # Retrieve the original column and use that for ordering
                            if name == "Col":
                                column = annotation.output_field.column
                            else:
                                column = annotation.col.output_field.column

                    field = query.model._meta.get_field(column)

                    if field.get_internal_type() in (u"TextField", u"BinaryField"):
                        raise NotSupportedError(INVALID_ORDERING_FIELD_MESSAGE)

                    # If someone orders by 'fk' rather than 'fk_id' this complains as that should take
                    # into account the related model ordering. Note the difference between field.name == column
                    # and field.attname (X_id)
                    if field.related_model and field.name == column and field.related_model._meta.ordering:
                        raise NotSupportedError("Related ordering is not supported on the Datastore")

                    column = "__key__" if field.primary_key else field.column
                    final.append("-" + column if col.startswith("-") else column)
                except FieldDoesNotExist:
                    if col in query.extra_select:
                        # If the column is in the extra select we transform to the original
                        # column
                        try:
                            field = opts.get_field(query.extra_select[col][0])
                            column = "__key__" if field.primary_key else field.column
                            final.append("-" + column if col.startswith("-") else column)
                            continue
                        except FieldDoesNotExist:
                            # Just pass through to the exception below
                            pass

                    available = opts.get_all_field_names()
                    raise FieldError(
                        "Cannot resolve keyword %r into field. " "Choices are: %s" % (col, ", ".join(available))
                    )

        # Reverse if not using standard ordering
        def swap(col):
            if col.startswith("-"):
                return col.lstrip("-")
            else:
                return "-{}".format(col)

        if not query.standard_ordering:
            final = [x if x in expressions else swap(x) for x in final]

        if len(final) != len(result):
            diff = set(result) - set(final)
            log_once(
                logger.warning,
                "The following orderings were ignored as cross-table and random orderings"
                " are not supported on the Datastore: %s",
                diff,
            )

        return final



from django.core import signals
from .caching import reset_context

POLYMODEL_CLASS_ATTRIBUTE = "class"


def reset_context_cache(**kwargs):
    reset_context()


# We connect to request_started in case something stopped request_finished
# being fired. We connect to request_finished to minimize the time (and risk)
# that something is left in memory
signals.request_started.connect(reset_context_cache)
signals.request_finished.connect(reset_context_cache)



"""
The Datastore does not provide database level constraints around uniqueness
(unlike relational a SQL database, where you can ensure uniqueness for a certain
column, or a combination of columns).

To mimic the ability to define these constraints using the Django API,
we have implemented an approach where, thanks to Cloud Firestore strong
consistency we check unique constraints transactionally before any write

This allows us to efficiently check for existing constraints before doing a put().
"""

from .dbapi import IntegrityError
from .unique_utils import (
    unique_identifiers_from_entity,
    _has_enabled_constraints,
    _has_unique_constraints,
)


UNIQUE_MARKER_KIND = "uniquemarker"
CONSTRAINT_VIOLATION_MSG = "Unique constraint violation for kind {} on fields: {}"


def has_active_unique_constraints(model_or_instance):
    """
    Returns a boolean to indicate if we should respect any unique constraints
    defined on the provided instance / model, taking into account any model
    or global related flags.
    """
    # are unique constraints disabled on the provided model take precident
    constraints_enabled = _has_enabled_constraints(model_or_instance)
    if not constraints_enabled:
        return False

    # does the object have unique constraints defined in the model definition
    return _has_unique_constraints(model_or_instance)


def check_unique_markers_in_memory(model, entities):
    """
    Compare the entities using their in memory properties, to see if any
    unique constraints are violated.

    This would always need to be used in conjunction with RPC checks against
    persisted markers to ensure data integrity.
    """
    all_unique_marker_key_values = set([])
    for entity, _ in entities:
        unique_marker_key_values = unique_identifiers_from_entity(model, entity, ignore_pk=True)
        for named_key in unique_marker_key_values:
            if named_key not in all_unique_marker_key_values:
                all_unique_marker_key_values.add(named_key)
            else:
                table_name = named_key.split("|")[0]
                unique_fields = named_key.split("|")[1:]
                raise IntegrityError(CONSTRAINT_VIOLATION_MSG.format(table_name, unique_fields))



from django.core.exceptions import ImproperlyConfigured
from django.db import models

from gcloudc.core import validators


_MAX_STRING_LENGTH = 1500


class CharOrNoneField(models.CharField):
    """ A field that stores only non-empty strings or None (it won't store empty strings).
        This is useful if you want values to be unique but also want to allow empty values.
    """

    empty_strings_allowed = False

    def __init__(self, *args, **kwargs):
        # Don't allow null=False because that would be insane.
        if not kwargs.get("null", True):
            raise ImproperlyConfigured("You can't set null=False on a CharOrNoneField.")
        # Set blank=True as the default, but allow it to be overridden, as it's theoretically
        # possible that you might want to prevent emptiness only in a form
        defaults = dict(null=True, blank=True, default=None)
        defaults.update(**kwargs)
        super(CharOrNoneField, self).__init__(*args, **defaults)

    def pre_save(self, model_instance, add):
        value = super(CharOrNoneField, self).pre_save(model_instance, add)
        # Change empty strings to None
        if not value:
            return None
        return value


class CharField(models.CharField):
    def __init__(self, max_length=_MAX_STRING_LENGTH, *args, **kwargs):
        assert max_length <= _MAX_STRING_LENGTH, "%ss max_length must not be greater than %d bytes." % (
            self.__class__.__name__,
            _MAX_STRING_LENGTH,
        )

        super(CharField, self).__init__(max_length=max_length, *args, **kwargs)

        # Append the MaxBytesValidator if it's not been included already
        self.validators = [x for x in self.validators if not isinstance(x, validators.MaxBytesValidator)] + [
            validators.MaxBytesValidator(limit_value=max_length)
        ]



import sleuth
from django import forms
from django.core import serializers
from django.core.exceptions import (
    ImproperlyConfigured,
    ValidationError,
)
from django.db.utils import IntegrityError
from gcloudc.db.models.fields.related import (
    RelatedListField,
    RelatedSetField,
)

from . import TestCase
from .models import (
    ISModel,
    ISOther,
    IterableFieldsWithValidatorsModel,
    IterableRelatedModel,
    IterableRelatedWithNonIntPkModel,
    ModelDatabaseA,
    ModelDatabaseB,
    RelatedListFieldRemoveDuplicatesModel,
    StringPkModel,
    RelatedCharFieldModel,
    Tag,
    GenericRelationModel,
    Post,
    RelationWithoutReverse,
    RelationWithOverriddenDbTable,
)


class RelatedListFieldRemoveDuplicatesForm(forms.ModelForm):
    class Meta:
        model = RelatedListFieldRemoveDuplicatesModel
        fields = ["related_list_field"]


class RelatedIterableFieldTests(TestCase):
    """ Combined tests for common RelatedListField and RelatedSetField tests. """

    multi_db = True

    def test_cannot_have_min_length_and_blank(self):
        """ Having min_length=X, blank=True doesn't make any sense, especially when you consider
            that django will skip the min_length check when the value (list/set)is empty.
        """
        self.assertRaises(ImproperlyConfigured, RelatedListField, ISModel, min_length=1, blank=True)
        self.assertRaises(ImproperlyConfigured, RelatedSetField, ISModel, min_length=1, blank=True)

    def test_related_list_field_set_field_min_max_lengths_valid(self):
        """ Test that when the min_legnth and max_length of a ListField and SetField are correct
            that no validation error is rasied.
        """
        others = []
        for x in range(2):
            others.append(ISOther.objects.create())
        instance = IterableFieldsWithValidatorsModel(
            list_field=["1", "2"],  # not being tested here
            set_field=set(["1", "2"]),  # not being tested here
            related_set=set(others),
            related_list=others,
        )
        instance.full_clean()

    def test_related_list_field_max_length_invalid(self):
        others = []
        for x in range(5):
            others.append(ISOther.objects.create())
        instance = IterableFieldsWithValidatorsModel(
            list_field=["1", "2"],  # not being tested here
            set_field=set(["1", "2"]),  # not being tested here
            related_set=set(others[:2]),  # not being tested here
            related_list=others,
        )
        self.assertRaisesMessage(
            ValidationError,
            "{'related_list': ['Ensure this field has at most 3 items (it has 5).']}",
            instance.full_clean,
        )

    def test_related_list_field_min_length_invalid(self):
        others = []
        for x in range(2):
            others.append(ISOther.objects.create())
        instance = IterableFieldsWithValidatorsModel(
            list_field=["1", "2"],  # not being tested here
            set_field=set(["1", "2"]),  # not being tested here
            related_set=set(others),
            related_list=others[:1],
        )
        self.assertRaisesMessage(
            ValidationError,
            "{'related_list': ['Ensure this field has at least 2 items (it has 1).']}",
            instance.full_clean,
        )

    def test_related_set_field_max_length_invalid(self):
        others = []
        for x in range(5):
            others.append(ISOther.objects.create())
        instance = IterableFieldsWithValidatorsModel(
            list_field=["1", "2"],  # not being tested here
            set_field=set(["1", "2"]),  # not being tested here
            related_list=others[:2],  # not being tested here
            related_set=set(others),
        )
        self.assertRaisesMessage(
            ValidationError,
            "{'related_set': ['Ensure this field has at most 3 items (it has 5).']}",
            instance.full_clean,
        )

    def test_related_set_field_min_length_invalid(self):
        others = []
        for x in range(2):
            others.append(ISOther.objects.create())
        instance = IterableFieldsWithValidatorsModel(
            list_field=["1", "2"],  # not being tested here
            set_field=set(["1", "2"]),  # not being tested here
            related_list=others,  # not being tested here
            related_set=set(others[:1]),
        )
        self.assertRaisesMessage(
            ValidationError,
            "{'related_set': ['Ensure this field has at least 2 items (it has 1).']}",
            instance.full_clean,
        )

    def test_model_stores_ids_as_integers_when_saving(self):
        others = []
        for x in range(2):
            others.append(ISOther.objects.create())

        instance = IterableRelatedModel(
            related_set_ids=[str(x.pk) for x in others], related_list_ids=[str(x.pk) for x in others]
        )

        instance.save()
        instance.refresh_from_db()

        self.assertEqual(instance.related_set_ids, [int(x.pk) for x in others])
        self.assertEqual(instance.related_list_ids, [int(x.pk) for x in others])

    def test_model_stores_ids_as_non_integers(self):
        others = []
        for x in range(2):
            others.append(StringPkModel.objects.create(name=str(x)))

        instance = IterableRelatedWithNonIntPkModel(
            related_set_ids=[x.pk for x in others], related_list_ids=[x.pk for x in others]
        )

        instance.save()
        instance.refresh_from_db()

        self.assertEqual(instance.related_set_ids, [x.pk for x in others])
        self.assertEqual(instance.related_list_ids, [x.pk for x in others])

    def test_related_set_field_cross_database(self):
        a = ModelDatabaseA.objects.create()
        b = ModelDatabaseB.objects.create()

        self.assertItemsEqual(a.set_of_bs.all(), [])

        a.set_of_bs.add(b)
        a.save()
        self.assertItemsEqual(a.set_of_bs.all(), [b])

        a.set_of_bs.remove(b)
        a.save()
        self.assertItemsEqual(a.set_of_bs.all(), [])

    def test_related_list_field_cross_database(self):
        a = ModelDatabaseA.objects.create()
        b = ModelDatabaseB.objects.create()

        self.assertItemsEqual(a.list_of_bs.all(), [])

        a.list_of_bs.add(b)
        a.save()
        self.assertItemsEqual(a.list_of_bs.all(), [b])

        a.list_of_bs.remove(b)
        a.save()
        self.assertItemsEqual(a.list_of_bs.all(), [])


class RelatedListFieldModelTests(TestCase):
    def test_values_list_queries_work(self):
        a, b = ISOther.objects.create(name="A"), ISOther.objects.create(name="B")
        thing = ISModel.objects.create(related_list_ids=[a.pk, b.pk])

        result = list(thing.related_list.values_list("pk"))
        self.assertEqual(result, [(a.pk,), (b.pk,)])

        result = list(thing.related_list.values_list("pk", flat=True))
        self.assertEqual(result, [a.pk, b.pk])

        result = list(thing.related_list.values_list("name", flat=True))
        self.assertEqual(result, ["A", "B"])

        result = list(thing.related_list.values_list("name"))
        self.assertEqual(result, [("A",), ("B",)])

        result = list(thing.related_list.values_list("pk", "name"))
        self.assertEqual(result, [(a.pk, "A"), (b.pk, "B")])

    def test_indexing_doesnt_over_fetch(self):
        a, b = ISOther.objects.create(), ISOther.objects.create()
        thing = ISModel.objects.create(related_list_ids=[a.pk, b.pk])

        with sleuth.watch("gcloudc.db.backends.datastore.meta_queries.QueryByKeys.__init__") as get:
            thing.related_list.all()[0]

            self.assertEqual(1, get.call_count)
            self.assertEqual(1, len(get.calls[0].args[3]))

    def test_can_update_related_field_from_form(self):
        related = ISOther.objects.create()
        thing = ISModel.objects.create(related_list=[related])
        before_list = thing.related_list
        thing.related_list.field.save_form_data(thing, [])
        self.assertNotEqual(before_list.all(), thing.related_list.all())

    def test_filtering_on_iterable_fields(self):
        related1 = ISOther.objects.create()
        related2 = ISOther.objects.create()
        related3 = ISOther.objects.create()
        thing = ISModel.objects.create(related_list=[related1, related2])
        # filtering using __contains lookup with ListField:
        qry = ISModel.objects.filter(related_list__contains=related1)
        self.assertEqual(sorted(x.pk for x in qry), [thing.pk])
        # filtering using __overlap lookup with ListField:
        qry = ISModel.objects.filter(related_list__overlap=[related2, related3])
        self.assertEqual(sorted(x.pk for x in qry), [thing.pk])

    def test_saving_forms(self):
        class TestForm(forms.ModelForm):
            class Meta:
                model = ISModel
                fields = ("related_list",)

        related = ISOther.objects.create()
        post_data = {"related_list": [str(related.pk)]}

        form = TestForm(post_data)
        self.assertTrue(form.is_valid())
        instance = form.save()
        self.assertEqual([related.pk], instance.related_list_ids)

    def test_remove_duplicates(self):
        """
        make sure that remove_duplicates option works fine for RelatedListField
        """
        instance_one, instance_two, instance_three, instance_four, instance_five = [
            RelatedCharFieldModel.objects.create(char_field=str(x)) for x in range(5)
        ]
        data = {
            "related_list_field": [
                instance_two.pk,
                instance_three.pk,
                instance_one.pk,
                instance_two.pk,
                instance_four.pk,
            ]
        }
        form = RelatedListFieldRemoveDuplicatesForm(data)
        self.assertTrue(form.is_valid())
        obj = form.save()

        self.assertEqual(
            obj.related_list_field_ids, [instance_two.pk, instance_three.pk, instance_one.pk, instance_four.pk]
        )
        obj.related_list_field.add(instance_four, instance_five)
        obj.save()
        self.assertEqual(
            obj.related_list_field_ids,
            [instance_two.pk, instance_three.pk, instance_one.pk, instance_four.pk, instance_five.pk],
        )


class RelatedSetFieldModelTests(TestCase):
    def test_can_update_related_field_from_form(self):
        related = ISOther.objects.create()
        thing = ISModel.objects.create(related_things={related})
        before_set = thing.related_things
        thing.related_list.field.save_form_data(thing, set())
        thing.save()
        self.assertNotEqual(before_set.all(), thing.related_things.all())

    def test_prefetch_related(self):
        tag = Tag.objects.create(name="Apples")

        for i in range(2):
            Post.objects.create(content="Bananas", tags={tag})

        posts = list(Post.objects.prefetch_related("tags").all())
        self.assertNumQueries(0, list, posts[0].tags.all())

    def test_saving_forms(self):
        class TestForm(forms.ModelForm):
            class Meta:
                model = ISModel
                fields = ("related_things",)

        related = ISOther.objects.create()
        post_data = {"related_things": [str(related.pk)]}

        form = TestForm(post_data)
        self.assertTrue(form.is_valid())
        instance = form.save()
        self.assertEqual({related.pk}, instance.related_things_ids)


class InstanceListFieldTests(TestCase):
    def test_deserialization(self):
        i1 = ISOther.objects.create(pk=1)
        i2 = ISOther.objects.create(pk=2)
        # Does the to_python need to return ordered list? SetField test only passes because the set
        # happens to order it correctly
        self.assertItemsEqual([i1, i2], ISModel._meta.get_field("related_list").to_python("[1, 2]"))

    def test_prefetch_related(self):
        tags = [Tag.objects.create(name="1"), Tag.objects.create(name="2"), Tag.objects.create(name="3")]

        # Extra one to make sure we're filtering properly
        Tag.objects.create(name="unused")

        for i in range(3):
            Post.objects.create(content="Bananas", ordered_tags=tags)

        with self.assertNumQueries(2):
            # 1 query on Posts + 1 query on Tags
            posts = list(Post.objects.prefetch_related("ordered_tags").all())

        with self.assertNumQueries(0):
            posts[0].tags.all()
            posts[1].tags.all()
            posts[2].tags.all()

        self.assertEqual(posts[0].ordered_tags.all()[0].name, "1")
        self.assertEqual(posts[0].ordered_tags.all()[1].name, "2")
        self.assertEqual(posts[0].ordered_tags.all()[2].name, "3")

    def test_default_on_delete_does_nothing(self):
        child = ISOther.objects.create(pk=1)
        parent = ISModel.objects.create(related_list=[child])

        child.delete()

        try:
            parent = ISModel.objects.get(pk=parent.pk)
            self.assertEqual([1], parent.related_list_ids)
        except ISModel.DoesNotExist:
            self.fail("Parent instance was deleted, apparently by on_delete=CASCADE")

    def test_save_and_load_empty(self):
        """
        Create a main object with no related items,
        get a copy of it back from the db and try to read items.
        """
        main = ISModel.objects.create()
        main_from_db = ISModel.objects.get(pk=main.pk)

        # Fetch the container from the database and read its items
        self.assertItemsEqual(main_from_db.related_list.all(), [])

    def test_basic_usage(self):
        main = ISModel.objects.create()
        other = ISOther.objects.create(name="test")
        other2 = ISOther.objects.create(name="test2")

        main.related_list.add(other)
        main.save()

        self.assertEqual([other.pk], main.related_list_ids)
        self.assertEqual(list(ISOther.objects.filter(pk__in=main.related_list_ids)), list(main.related_list.all()))
        self.assertEqual([main], list(other.ismodel_list.all()))

        main.related_list.remove(other)
        self.assertFalse(main.related_list)

        main.related_list = [other2]
        self.assertEqual([other2.pk], main.related_list_ids)

        with self.assertRaises(AttributeError):
            other.ismodel_list = [main]

        without_reverse = RelationWithoutReverse.objects.create(name="test3")
        self.assertFalse(hasattr(without_reverse, "ismodel_list"))

    def test_add_to_empty(self):
        """
        Create a main object with no related items,
        get a copy of it back from the db and try to add items.
        """
        main = ISModel.objects.create()
        main_from_db = ISModel.objects.get(pk=main.pk)

        other = ISOther.objects.create()
        main_from_db.related_list.add(other)
        main_from_db.save()

    def test_add_another(self):
        """
        Create a main object with related items,
        get a copy of it back from the db and try to add more.
        """
        main = ISModel.objects.create()
        other1 = ISOther.objects.create()
        main.related_things.add(other1)
        main.save()

        main_from_db = ISModel.objects.get(pk=main.pk)
        other2 = ISOther.objects.create()

        main_from_db.related_list.add(other2)
        main_from_db.save()

    def test_multiple_objects(self):
        main = ISModel.objects.create()
        other1 = ISOther.objects.create()
        other2 = ISOther.objects.create()

        main.related_list.add(other1, other2)
        main.save()

        main_from_db = ISModel.objects.get(pk=main.pk)
        self.assertEqual(main_from_db.related_list.count(), 2)

    def test_deletion(self):
        """
        Delete one of the objects referred to by the related field
        """
        main = ISModel.objects.create()
        other = ISOther.objects.create()
        main.related_list.add(other)
        main.save()

        other.delete()
        self.assertEqual(main.related_list.count(), 0)

    def test_ordering_is_maintained(self):
        main = ISModel.objects.create()
        other = ISOther.objects.create()
        other1 = ISOther.objects.create()
        other2 = ISOther.objects.create()
        other3 = ISOther.objects.create()
        main.related_list.add(other, other1, other2, other3)
        main.save()
        self.assertEqual(main.related_list.count(), 4)
        self.assertEqual([other.pk, other1.pk, other2.pk, other3.pk], main.related_list_ids)
        self.assertItemsEqual([other, other1, other2, other3], main.related_list.all())
        main.related_list.clear()
        main.save()
        self.assertEqual([], main.related_list_ids)

    def test_duplicates_maintained(self):
        """
            For whatever reason you might want many of the same relation in the
            list
        """
        main = ISModel.objects.create()
        other = ISOther.objects.create()
        other1 = ISOther.objects.create()
        other2 = ISOther.objects.create()
        other3 = ISOther.objects.create()
        main.related_list.add(other, other1, other2, other1, other3)
        main.save()
        self.assertEqual([other.pk, other1.pk, other2.pk, other1.pk, other3.pk], main.related_list_ids)
        self.assertItemsEqual([other, other1, other2, other1, other3], main.related_list.all())

    def test_slicing(self):
        main = ISModel.objects.create()
        other = ISOther.objects.create()
        other1 = ISOther.objects.create()
        other2 = ISOther.objects.create()
        other3 = ISOther.objects.create()
        main.related_list.add(other, other1, other2, other1, other3)
        main.save()
        self.assertItemsEqual([other, other1], main.related_list.all()[:2])
        self.assertItemsEqual([other1], main.related_list.all()[1:2])
        self.assertEqual(other1, main.related_list.all()[1:2][0])

    def test_filtering(self):
        main = ISModel.objects.create()
        other = ISOther.objects.create(name="one")
        other1 = ISOther.objects.create(name="two")
        other2 = ISOther.objects.create(name="one")
        ISOther.objects.create(name="three")
        main.related_list.add(other, other1, other2, other1, other2)
        main.save()
        self.assertItemsEqual([other, other2, other2], main.related_list.filter(name="one"))

    def test_related_list_field_serializes_and_deserializes(self):
        obj = ISModel.objects.create()
        foo = ISOther.objects.create(name="foo")
        bar = ISOther.objects.create(name="bar")
        obj.related_list.add(foo, bar)
        obj.save()

        data = serializers.serialize("json", [obj])
        new_obj = next(serializers.deserialize("json", data)).object
        self.assertEqual(list(new_obj.related_list.all()), [foo, bar])


class InstanceSetFieldTests(TestCase):
    def test_deserialization(self):
        i1 = ISOther.objects.create(pk=1)
        i2 = ISOther.objects.create(pk=2)

        self.assertEqual(set([i1, i2]), ISModel._meta.get_field("related_things").to_python("[1, 2]"))

    def test_basic_usage(self):
        main = ISModel.objects.create()
        other = ISOther.objects.create(name="test")
        other2 = ISOther.objects.create(name="test2")

        main.related_things.add(other)
        main.save()

        self.assertEqual({other.pk}, main.related_things_ids)
        self.assertEqual(list(ISOther.objects.filter(pk__in=main.related_things_ids)), list(main.related_things.all()))

        self.assertItemsEqual([main], ISModel.objects.filter(related_things__contains=other).all())
        self.assertItemsEqual([main], list(other.ismodel_set.all()))

        main.related_things.remove(other)
        self.assertFalse(main.related_things_ids)

        main.related_things = {other2}
        self.assertEqual({other2.pk}, main.related_things_ids)

        with self.assertRaises(AttributeError):
            other.ismodel_set = {main}

        without_reverse = RelationWithoutReverse.objects.create(name="test3")
        self.assertFalse(hasattr(without_reverse, "ismodel_set"))

    def test_save_and_load_empty(self):
        """
        Create a main object with no related items,
        get a copy of it back from the db and try to read items.
        """
        main = ISModel.objects.create()
        main_from_db = ISModel.objects.get(pk=main.pk)

        # Fetch the container from the database and read its items
        self.assertItemsEqual(main_from_db.related_things.all(), [])

    def test_add_to_empty(self):
        """
        Create a main object with no related items,
        get a copy of it back from the db and try to add items.
        """
        main = ISModel.objects.create()
        main_from_db = ISModel.objects.get(pk=main.pk)

        other = ISOther.objects.create()
        main_from_db.related_things.add(other)
        main_from_db.save()

    def test_add_another(self):
        """
        Create a main object with related items,
        get a copy of it back from the db and try to add more.
        """
        main = ISModel.objects.create()
        other1 = ISOther.objects.create()
        main.related_things.add(other1)
        main.save()

        main_from_db = ISModel.objects.get(pk=main.pk)
        other2 = ISOther.objects.create()

        main_from_db.related_things.add(other2)
        main_from_db.save()

    def test_multiple_objects(self):
        main = ISModel.objects.create()
        other1 = ISOther.objects.create()
        other2 = ISOther.objects.create()

        main.related_things.add(other1, other2)
        main.save()

        main_from_db = ISModel.objects.get(pk=main.pk)
        self.assertEqual(main_from_db.related_things.count(), 2)

    def test_deletion(self):
        """
        Delete one of the objects referred to by the related field
        """
        main = ISModel.objects.create()
        other = ISOther.objects.create()
        main.related_things.add(other)
        main.save()

        other.delete()
        self.assertEqual(main.related_things.count(), 0)

    def test_querying_with_isnull(self):
        obj = ISModel.objects.create()

        self.assertItemsEqual([obj], ISModel.objects.filter(related_things__isempty=True))
        self.assertItemsEqual([obj], ISModel.objects.filter(related_things_ids__isempty=True))

    def test_related_set_field_serializes_and_deserializes(self):
        obj = ISModel.objects.create()
        foo = ISOther.objects.create(name="foo")
        bar = ISOther.objects.create(name="bar")
        obj.related_things.add(foo, bar)
        obj.save()

        data = serializers.serialize("json", [obj])

        new_obj = next(serializers.deserialize("json", data)).object
        self.assertEqual(set(new_obj.related_things.all()), set([foo, bar]))


class TestGenericRelationField(TestCase):
    def test_basic_usage(self):
        instance = GenericRelationModel.objects.create()
        self.assertIsNone(instance.relation_to_anything)

        thing = ISOther.objects.create()
        instance.relation_to_anything = thing
        instance.save()

        self.assertTrue(instance.relation_to_anything_id)

        instance = GenericRelationModel.objects.get()
        self.assertEqual(thing, instance.relation_to_anything)

    def test_overridden_dbtable(self):
        """ Check that the related object having a custom `db_table` doesn't affect the functionality. """
        instance = GenericRelationModel.objects.create()
        self.assertIsNone(instance.relation_to_anything)

        weird = RelationWithOverriddenDbTable.objects.create()
        instance.relation_to_anything = weird
        instance.save()

        self.assertTrue(instance.relation_to_anything)

        instance = GenericRelationModel.objects.get()
        self.assertEqual(weird, instance.relation_to_anything)

    def test_querying(self):
        thing = ISOther.objects.create()
        instance = GenericRelationModel.objects.create(relation_to_anything=thing)
        self.assertEqual(GenericRelationModel.objects.filter(relation_to_anything=thing)[0], instance)

    def test_unique(self):
        thing = ISOther.objects.create()
        instance = GenericRelationModel.objects.create(unique_relation_to_anything=thing)
        # Trying to create another instance which relates to the same 'thing' should fail
        self.assertRaises(IntegrityError, GenericRelationModel.objects.create, unique_relation_to_anything=thing)
        # But creating 2 objects which both have `unique_relation_to_anything` set to None should be fine
        instance.unique_relation_to_anything = None
        instance.save()
        GenericRelationModel.objects.create(unique_relation_to_anything=None)
        GenericRelationModel.objects.create()  # It should work even if we don't explicitly set it to None

    def test_saving_forms(self):
        class TestForm(forms.ModelForm):
            class Meta:
                model = GenericRelationModel
                fields = ("relation_to_anything",)

        related = ISOther.objects.create()
        post_data = {"relation_to_anything_0": related.__class__._meta.db_table, "relation_to_anything_1": related.pk}

        form = TestForm(post_data)
        self.assertTrue(form.is_valid())
        instance = form.save()
        self.assertEqual(related, instance.relation_to_anything)



from django.test import TestCase as DjangoTestCase
from django.db import connection


class TestCase(DjangoTestCase):
    def setUp(self):
        connection.connect()  # Make sure we have a connection
        super().setUp()

    # This was mistakenly renamed to assertCountsEqual
    # in Python 3, so this avoids any complications arising
    # when they rectify that! https://bugs.python.org/issue27060
    def assertItemsEqual(self, lhs, rhs):
        if set(lhs) != set(rhs):
            raise AssertionError("Items were not the same in both lists")



import copy
from functools import cmp_to_key
from itertools import product

from django.conf import settings
from django.db import NotSupportedError
from django.core.exceptions import EmptyResultSet
from google.cloud.datastore.key import Key

from .query import WhereNode
from .query_utils import compare_keys

# Maximum number of subqueries in a multiquery
DEFAULT_MAX_ALLOWABLE_QUERIES = 100


def preprocess_node(node, negated):

    to_remove = []

    # Go through the children of this node and if any of the
    # child nodes are leaf nodes, then explode them if necessary
    for child in node.children:
        if child.is_leaf:
            if child.operator == "ISNULL":
                value = not child.value if node.negated else child.value
                if value:
                    child.operator = "="
                    child.value = None
                else:
                    child.operator = ">"
                    child.value = None

            elif node.negated and child.operator == "=":
                # Excluded equalities become inequalities

                lhs, rhs = WhereNode(node.using), WhereNode(node.using)
                lhs.column = rhs.column = child.column
                lhs.value = rhs.value = child.value
                lhs.operator = "<"
                rhs.operator = ">"

                child.operator = child.value = child.column = None
                child.connector = "OR"
                child.children = [lhs, rhs]

                assert not child.is_leaf

            elif child.operator == "IN":
                # Explode IN filters into a series of 'OR statements to make life
                # easier later
                new_children = []

                for value in child.value:
                    if node.negated:
                        lhs, rhs = WhereNode(node.using), WhereNode(node.using)
                        lhs.column = rhs.column = child.column
                        lhs.value = rhs.value = value
                        lhs.operator = "<"
                        rhs.operator = ">"

                        bridge = WhereNode(node.using)
                        bridge.connector = "OR"
                        bridge.children = [lhs, rhs]

                        new_children.append(bridge)
                    else:
                        new_node = WhereNode(node.using)
                        new_node.operator = "="
                        new_node.value = value
                        new_node.column = child.column
                        new_children.append(new_node)

                child.column = None
                child.operator = None
                child.connector = "AND" if negated else "OR"
                child.value = None
                child.children = new_children

                assert not child.is_leaf

            elif child.operator == "RANGE":
                lhs, rhs = WhereNode(node.using), WhereNode(node.using)
                lhs.column = rhs.column = child.column
                if node.negated:
                    lhs.operator = "<"
                    rhs.operator = ">"
                    child.connector = "OR"
                else:
                    lhs.operator = ">="
                    rhs.operator = "<="
                    child.connector = "AND"
                lhs.value = child.value[0]
                rhs.value = child.value[1]

                child.column = child.operator = child.value = None
                child.children = [lhs, rhs]

                assert not child.is_leaf
        elif node.negated:
            # Move the negation down the tree
            child.negated = not child.negated

    # If this node was negated, we flip everything
    if node.negated:
        node.negated = False
        node.connector = "AND" if node.connector == "OR" else "OR"

    for child in to_remove:
        node.children.remove(child)

    return node


def normalize_query(query):
    where = query.where

    # If there are no filters then this is already normalized
    if where is None:
        return query

    def walk_tree(where, original_negated=False):
        negated = original_negated

        if where.negated:
            negated = not negated

        preprocess_node(where, negated)

        rewalk = False
        for child in where.children:
            if where.connector == "AND" and child.children and child.connector == "AND" and not child.negated:
                where.children.remove(child)
                where.children.extend(child.children)
                rewalk = True
            elif child.connector == "AND" and len(child.children) == 1 and not child.negated:
                # Promote leaf nodes if they are the only child under an AND. Just for consistency
                where.children.remove(child)
                where.children.extend(child.children)
                rewalk = True
            elif len(child.children) > 1 and child.connector == "AND" and child.negated:
                new_grandchildren = []
                for grandchild in child.children:
                    new_node = WhereNode(child.using)
                    new_node.negated = True
                    new_node.children = [grandchild]
                    new_grandchildren.append(new_node)
                child.children = new_grandchildren
                child.connector = "OR"
                rewalk = True
            else:
                walk_tree(child, negated)

        if rewalk:
            walk_tree(where, original_negated)

        if where.connector == "AND" and any([x.connector == "OR" for x in where.children]):
            # ANDs should have been taken care of!
            assert not any([x.connector == "AND" and not x.is_leaf for x in where.children])

            product_list = []
            for child in where.children:
                if child.connector == "OR":
                    product_list.append(child.children)
                else:
                    product_list.append([child])

            producted = product(*product_list)

            new_children = []
            for branch in producted:
                new_and = WhereNode(where.using)
                new_and.connector = "AND"
                new_and.children = list(copy.deepcopy(branch))
                new_children.append(new_and)

            where.connector = "OR"
            where.children = list(set(new_children))
            walk_tree(where, original_negated)

        elif where.connector == "OR":
            new_children = []
            for child in where.children:
                if child.connector == "OR":
                    new_children.extend(child.children)
                else:
                    new_children.append(child)
            where.children = list(set(new_children))

    walk_tree(where)

    if where.connector != "OR":
        new_node = WhereNode(where.using)
        new_node.connector = "OR"
        new_node.children = [where]
        query._where = new_node

    all_pks = True
    for and_branch in query.where.children:
        if and_branch.is_leaf:
            children = [and_branch]
        else:
            children = and_branch.children

        for node in children:
            if node.column == "__key__" and node.operator in ("=", "IN"):
                break
        else:
            all_pks = False
            break

    MAX_ALLOWABLE_QUERIES = getattr(settings, "DJANGAE_MAX_QUERY_BRANCHES", DEFAULT_MAX_ALLOWABLE_QUERIES)

    if (not all_pks) and len(query.where.children) > MAX_ALLOWABLE_QUERIES:
        raise NotSupportedError(
            ("Unable to run query as it required more than {} subqueries "
             "(limit is configurable with DJANGAE_MAX_QUERY_BRANCHES)").format(
                MAX_ALLOWABLE_QUERIES
            )
        )

    def remove_empty_in(node):
        """
            Once we are normalized, if any of the branches filters
            on an empty list, we can remove that entire branch from the
            query. If this leaves no branches, then the result set is empty
        """

        # This is a bit ugly, but you try and do it more succinctly :)
        # We have the following possible situations for IN queries with an empty
        # value:

        # - Negated: One of the nodes in the and branch will always be true and is therefore
        #    unnecessary, we leave it alone though
        # - Not negated: The entire AND branch will always be false, so that branch can be removed
        #    if that was the last branch, then the queryset will be empty

        # Everything got wiped out!
        if node.connector == "OR" and len(node.children) == 0:
            raise EmptyResultSet()

        for and_branch in node.children[:]:
            if and_branch.is_leaf and and_branch.operator == "IN" and not len(and_branch.value):
                node.children.remove(and_branch)

            if not node.children:
                raise EmptyResultSet()

    remove_empty_in(where)

    def detect_conflicting_key_filter(node):
        assert node.connector == "OR"
        for and_branch in node.children[:]:
            # If we have a Root OR with leaf elements, we don't need to worry
            if and_branch.is_leaf:
                break

            pk_equality_found = None
            for child in and_branch.children:
                if child.column == "__key__" and child.operator == "=":
                    if pk_equality_found and pk_equality_found != child.value:
                        # Remove this AND branch as it's impossible to return anything
                        if and_branch in node.children:
                            node.children.remove(and_branch)
                    else:
                        pk_equality_found = child.value
            if not node.children:
                raise EmptyResultSet()

    detect_conflicting_key_filter(query.where)

    def remove_unnecessary_nodes(top_node):
        """
            Sometimes you end up with a branch that has two nodes that
            have the same column and operator, but different values. When
            this happens we need to simplify, e.g.:

            AND:[username<A],[username<B] -> AND:[username<A]
        """

        for and_branch in top_node.children[:]:
            seen = {}

            altered = False
            for node in and_branch.children:
                key = (node.column, node.operator)
                if key in seen:
                    altered = True

                    cmp_kwargs = {}
                    if isinstance(seen[key].value, Key) or isinstance(node.value, Key):
                        cmp_kwargs["key"] = cmp_to_key(compare_keys)

                    if node.operator in ('<', '<='):
                        seen[key].value = min(seen[key].value, node.value, **cmp_kwargs)
                    elif node.operator in ('>', '>='):
                        seen[key].value = max(seen[key].value, node.value, **cmp_kwargs)
                    elif node.operator == "=":
                        # Impossible filter! remove the AND branch entirely
                        if and_branch in top_node.children and seen[key].value != node.value:
                            top_node.children.remove(and_branch)
                        break
                    else:
                        pass
                else:
                    seen[key] = node

            if altered:
                and_branch.children = [x for x in seen.values()]

        # If all the OR clause are impossible filters we end up with no filters
        # at all, which is incorrect
        if not top_node.children:
            raise EmptyResultSet()

    remove_unnecessary_nodes(query.where)

    return query



import copy
import threading
import uuid

from google.cloud import exceptions
from google.cloud.datastore.transaction import \
    Transaction as DatastoreTransaction

from django.db import connections, DatabaseError
from gcloudc import context_decorator
from gcloudc.db.backends.datastore import caching

TRANSACTION_ENTITY_LIMIT = 500


def in_atomic_block(using="default"):
    txn = current_transaction()
    if not txn:
        return False

    datastore_transaction = getattr(txn, "_datastore_transaction")
    if not datastore_transaction:
        return False

    # Returns True if this is a Datastore transaction, and not a normal "Batch"
    return isinstance(datastore_transaction, DatastoreTransaction)


class Transaction(object):
    def __init__(self, connection, datastore_transaction=None):
        self._connection = connection
        self._datastore_transaction = datastore_transaction
        self._seen_keys = set()

    def _generate_id(self):
        """
            The Datastore API won't generate keys automatically until a
            transaction commits, that's too late!

            This is a hack, it might be the only hack we can do :(

            Note. Even though the Datastore can handle negative IDs (it's a signed
            64 bit integer) the default allocate never does, and also, this breaks
            Django URL regexes etc. So like the allocator we just do 32 bit ones.
        """
        unsigned = uuid.uuid4().int & (1 << 32) - 1
        return unsigned

    def key(self, *args, **kwargs):
        """
        Interface to the Datastore client Key factory.
        """
        assert("namespace" not in kwargs)
        namespace = self._connection.namespace
        kwargs["namespace"] = namespace

        parent = kwargs.get("parent")
        if parent:
            assert(parent.namespace == namespace)

        return self._connection.gclient.key(*args, **kwargs)

    def get(self, key_or_keys, missing=None):
        # For some reason Datastore Transactions don't provide their
        # own get
        if hasattr(key_or_keys, "__iter__") and not isinstance(key_or_keys, str):
            getter = self._connection.gclient.get_multi
            ret = getter(key_or_keys, missing=missing)
            if ret:
                [self._seen_keys.add(x.key) for x in ret]
                return ret
        else:
            ret = self._connection.gclient.get(key_or_keys)
            if ret:
                self._seen_keys.add(ret.key)

        return ret

    def put_multi(self, entities):
        for entity in entities:
            self.put(entity)

    def put(self, entity):
        putter = self._datastore_transaction.put if self._datastore_transaction else self._connection.gclient.put

        putter(entity)

        assert entity.key

        self._seen_keys.add(entity.key)

        return entity.key

    def delete(self, key_or_keys):
        """
        Delete an entity or entities using a different API depending if we
        are currently in a transaction batch or not.
        """
        # if we've got an iterable of keys....
        if hasattr(key_or_keys, "__iter__"):
            # there is no delete_multi on the transaction object directly
            if self._datastore_transaction:
                for key in key_or_keys:
                    self._datastore_transaction.delete(key)
            else:
                self._connection.gclient.delete_multi(key_or_keys)
        else:
            if self._datastore_transaction:
                self._datastore_transaction.delete(key_or_keys)
            else:
                # delete() is just a wrapper around delete_multi anyway...
                self._connection.gclient.delete_multi([key_or_keys])

    def query(self, *args, **kwargs):
        return self._connection.gclient.query(*args, **kwargs)

    def enter(self):
        self._seen_keys = set()
        self._enter()

    def exit(self):
        self._exit()
        self._seen_keys = set()

    def _enter(self):
        raise NotImplementedError()

    def _exit(self):
        raise NotImplementedError()

    def has_already_been_read(self, instance):
        if instance.pk is None:
            return False

        if not self._datastore_transaction:
            return False

        key = self.key(instance._meta.db_table, instance.pk)
        return key in self._seen_keys

    def refresh_if_unread(self, instance):
        """
            Calls instance.refresh_from_db() if the instance hasn't already
            been read this transaction. This helps prevent oddities if you
            call nested transactional functions. e.g.

            @atomic()
            def my_method(self):
                self.refresh_from_db()   # << Refresh state from the start of the transaction
                self.update()
                self.save()

            with atomic():
                instance = MyModel.objects.get(pk=1)
                instance.other_update()
                instance.my_method()  # << Oops! Undid work!
                instance.save()

            Instead, this will fix it

            def my_method(self):
                with atomic() as txn:
                    txn.refresh_if_unread(self)
                    self.update()
                    self.save()
        """

        if self.has_already_been_read(instance):
            # If the instance has already been read this transaction,
            # then don't refresh it again.
            return
        else:
            instance.refresh_from_db()

    def _commit(self):
        if self._transaction:
            return self._transaction.commit()

    def _rollback(self):
        if self._transaction:
            self._transaction.rollback()


class IndependentTransaction(Transaction):
    def __init__(self, connection):
        txn = connection.gclient.transaction()
        super().__init__(connection, txn)

        self.owner = connection.ops.connection
        self.previous_on_commit = []

    def _enter(self):
        self.previous_on_commit = self.owner.run_on_commit
        self.owner.run_on_commit = []
        self._datastore_transaction.begin()

    def _exit(self):
        self._datastore_transaction = None
        self.owner.run_on_commit = self.previous_on_commit


class NestedTransaction(Transaction):
    def _enter(self):
        pass

    def _exit(self):
        pass


class NormalTransaction(Transaction):
    def __init__(self, connection):
        txn = connection.gclient.transaction()
        super().__init__(connection, txn)

    def _enter(self):
        self._datastore_transaction.begin()

    def _exit(self):
        self._datastore_transaction = None


class NoTransaction(Transaction):
    def _enter(self):
        pass

    def _exit(self):
        self._datastore_transaction = None


_STORAGE = threading.local()


def _rpc(using):
    """
        In the low-level connector code, we use this function
        to return a transaction to perform a Get/Put/Delete on
        this effectively returns the current_transaction or a new
        RootTransaction() which is basically no transaction at all.
    """

    assert(isinstance(using, str))
    assert(using)

    class RootTransaction(Transaction):
        def _enter(self):
            pass

        def _exit(self):
            pass

    return current_transaction(using) or RootTransaction(connections[using].connection)


def current_transaction(using="default"):
    """
        Returns the current 'Transaction' object (which may be a NoTransaction). This is useful
        when atomic() is used as a decorator rather than a context manager. e.g.

        @atomic()
        def my_function(apple):
            current_transaction().refresh_if_unread(apple)
            apple.thing = 1
            apple.save()
    """

    _init_storage()

    active_transaction = None

    # Return the last Transaction object with a connection
    for txn in reversed(_STORAGE.transaction_stack.get(using, [])):
        if isinstance(txn, IndependentTransaction):
            active_transaction = txn
            break
        elif isinstance(txn, NormalTransaction):
            active_transaction = txn
            # Keep searching... there may be an independent or further transaction
        elif isinstance(txn, NoTransaction):
            # Bail immediately for non_atomic blocks. There is no transaction there.
            active_transaction = None
            break

    return active_transaction


def _init_storage():
    if not hasattr(_STORAGE, "transaction_stack"):
        _STORAGE.transaction_stack = {}


class TransactionFailedError(DatabaseError):
    pass


class AtomicDecorator(context_decorator.ContextDecorator):
    """
    Exposes a decorator based API for transaction use. This in turn allows us
    to define the expected behaviour of each transaction via kwargs.

    For example passing `independent` creates a new transaction instance using
    the Datastore client under the hood. This is useful to workaround the
    limitations of 500 entity writes per transaction/batch.
    """

    VALID_ARGUMENTS = ("independent", "mandatory", "using", "read_only", "enable_cache")

    @classmethod
    def _do_enter(cls, state, decorator_args):
        _init_storage()

        mandatory = decorator_args.get("mandatory", False)
        independent = decorator_args.get("independent", False)
        read_only = decorator_args.get("read_only", False)
        using = decorator_args.get("using", "default")

        mandatory = False if mandatory is None else mandatory
        independent = False if independent is None else independent
        read_only = False if read_only is None else read_only
        state.using = using = "default" if using is None else using

        enable_cache = decorator_args.get("enable_cache", True)
        context = caching.get_context()
        state.original_context_enabled = context.context_enabled

        # Only force-disable cache, do not force enable (e.g. if already disabled
        # in the context, leave as is)
        if enable_cache is False:
            context.context_enabled = enable_cache

        new_transaction = None

        connection = connections[using]

        # Connect if necessary (mainly in tests)
        if not connection.connection:
            connection.connect()

        connection = connection.connection
        assert(connection)

        if independent:
            new_transaction = IndependentTransaction(connection)
        elif in_atomic_block():
            new_transaction = NestedTransaction(connection)
        elif mandatory:
            raise TransactionFailedError(
                "You've specified that an outer transaction is mandatory, but one doesn't exist"
            )
        else:
            new_transaction = NormalTransaction(connection)

        try:
            new_transaction.enter()
        except Exception:
            new_transaction.exit()
            raise
        else:
            _STORAGE.transaction_stack.setdefault(using, []).append(new_transaction)

            if isinstance(new_transaction, (IndependentTransaction, NormalTransaction)):
                caching.get_context().stack.push()

            # We may have created a new transaction, we may not. current_transaction() returns
            # the actual active transaction (highest NormalTransaction or lowest IndependentTransaction)
            # or None if we're in a non_atomic, or there are no transactions
            return current_transaction()

    @classmethod
    def _do_exit(cls, state, decorator_args, exception):
        _init_storage()

        connection = connections[state.using]
        transaction = _STORAGE.transaction_stack[state.using].pop()

        try:
            if transaction._datastore_transaction:
                if exception or connection.needs_rollback:
                    transaction._datastore_transaction.rollback()
                else:
                    try:
                        transaction._datastore_transaction.commit()

                        if isinstance(transaction, (IndependentTransaction, NormalTransaction)):
                            with non_atomic(using=state.using):
                                # Run Django commit hooks (if any)
                                connection.run_and_clear_commit_hooks()

                    except exceptions.GoogleCloudError:
                        raise TransactionFailedError()
        finally:
            if isinstance(transaction, (IndependentTransaction, NormalTransaction)):
                context = caching.get_context()
                # Clear the context cache at the end of a transaction
                if exception:
                    context.stack.pop(discard=True)
                else:
                    context.stack.pop(apply_staged=True, clear_staged=True)

                context.context_enabled = state.original_context_enabled

            if exception:
                connection.run_on_commit = []

            transaction.exit()


atomic = AtomicDecorator
commit_on_success = AtomicDecorator  # Alias to the old Django name for this kinda thing


class NonAtomicDecorator(AtomicDecorator):
    VALID_ARGUMENTS = ("using",)

    @classmethod
    def _do_enter(cls, state, decorator_args):
        _init_storage()

        context = caching.get_context()

        state.using = using = decorator_args.get("using", "default")
        connection = connections[using]

        # Connect if necessary (mainly in tests)
        if not connection.connection:
            connection.connect()

        connection = connection.connection
        assert(connection)

        # For non_atomic blocks we pass a Batch as the transaction
        new_transaction = NoTransaction(connection.gclient.batch())

        _STORAGE.transaction_stack.setdefault(using, []).append(new_transaction)
        _STORAGE.transaction_stack[using][-1].enter()

        # Store the current state of the stack (aside from the first entry)
        state.original_stack = copy.deepcopy(context.stack.stack[1:])

        # Unwind the in-context stack leaving just the first entry
        while len(context.stack.stack) > 1:
            context.stack.pop(discard=True)

        return current_transaction()

    @classmethod
    def _do_exit(cls, state, decorator_args, exception):
        _init_storage()

        context = caching.get_context()
        transaction = _STORAGE.transaction_stack[state.using].pop()

        try:
            assert(not transaction._datastore_transaction)
        finally:
            # Restore the context stack as it was
            context.stack.stack = context.stack.stack + state.original_stack
            transaction.exit()


non_atomic = NonAtomicDecorator



import datetime
import decimal
import logging
import os
import uuid
import warnings

import requests

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.backends.base.base import BaseDatabaseWrapper
from django.db.backends.base.client import BaseDatabaseClient
from django.db.backends.base.creation import BaseDatabaseCreation
from django.db.backends.base.features import BaseDatabaseFeatures
from django.db.backends.base.introspection import (
    BaseDatabaseIntrospection,
    TableInfo,
)
from django.db.backends.base.operations import BaseDatabaseOperations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.backends.base.validation import BaseDatabaseValidation
from django.utils import (
    timezone,
)
from django.utils.encoding import smart_text
from google.cloud import (
    datastore,
    environment_vars,
)

from . import dbapi as Database
from .commands import (
    DeleteCommand,
    FlushCommand,
    InsertCommand,
    SelectCommand,
    UpdateCommand,
    coerce_unicode,
)
from .indexing import load_special_indexes
from .utils import (
    decimal_to_string,
    ensure_datetime,
    get_datastore_key,
    make_utc_compatible,
)

logger = logging.getLogger(__name__)


class Connection(object):
    """ Dummy connection class """

    def __init__(self, wrapper, params):
        self.creation = wrapper.creation
        self.ops = wrapper.ops
        self.queries = []
        self.settings_dict = params
        self.namespace = wrapper.namespace

        self.gclient = datastore.Client(
            namespace=wrapper.namespace,
            project=params["PROJECT"],
            # avoid a bug in the google client - it tries to authenticate even when the emulator is enabled
            # see https://github.com/googleapis/google-cloud-python/issues/5738
            _http=requests.Session if os.environ.get(environment_vars.GCD_HOST) else None,
        )

    def acquire_constraint_markers(self, markers):
        pass

    def release_constraint_markers(self, markers):
        pass

    def rollback(self):
        pass

    def commit(self):
        pass

    def close(self):
        pass


class Cursor(object):
    """ Dummy cursor class """

    def __init__(self, connection):
        self.connection = connection
        self.start_cursor = None
        self.returned_ids = []
        self.rowcount = -1
        self.last_select_command = None
        self.last_delete_command = None

    def execute(self, sql, *params):
        if isinstance(sql, SelectCommand):
            # Also catches subclasses of SelectCommand (e.g Update)
            self.last_select_command = sql
            self.rowcount = self.last_select_command.execute() or -1
        elif isinstance(sql, FlushCommand):
            sql.execute()
        elif isinstance(sql, UpdateCommand):
            self.rowcount = sql.execute()
        elif isinstance(sql, DeleteCommand):
            self.rowcount = sql.execute()
        elif isinstance(sql, InsertCommand):
            self.connection.queries.append(sql)
            self.returned_ids = sql.execute()
        else:
            raise Database.CouldBeSupportedError(
                "Can't execute traditional SQL: '%s' (although perhaps we could make GQL work)" % sql
            )

    def next(self):
        row = self.fetchone()
        if row is None:
            raise StopIteration
        return row

    def fetchone(self, delete_flag=False):
        try:
            result = next(self.last_select_command.results)

            if isinstance(result, int):
                return (result,)

            query = self.last_select_command.query

            row = []

            # Prepend extra select values to the resulting row
            for col, select in query.extra_selects:
                row.append(result.get(col))

            for col in self.last_select_command.query.init_list:
                row.append(result.get(col))

            self.returned_ids.append(result.key.id_or_name)
            return row
        except StopIteration:
            return None

    def fetchmany(self, size, delete_flag=False):
        if not self.last_select_command.results:
            return []

        result = []
        for i in range(size):
            entity = self.fetchone(delete_flag)
            if entity is None:
                break

            # Python DB API suggests a list of tuples, and returning
            # a list-of-lists breaks some tests
            result.append(tuple(entity))

        return result

    @property
    def lastrowid(self):
        return self.returned_ids[-1].id_or_name

    def __iter__(self):
        return self

    def close(self):
        pass


MAXINT = 9223372036854775808

_BULK_BATCH_SIZE_SETTING = "BULK_BATCH_SIZE"


class DatabaseOperations(BaseDatabaseOperations):
    compiler_module = "gcloudc.db.backends.datastore.compiler"

    # Datastore will store all integers as 64bit long values
    integer_field_ranges = {
        "SmallIntegerField": (-MAXINT, MAXINT - 1),
        "IntegerField": (-MAXINT, MAXINT - 1),
        "BigIntegerField": (-MAXINT, MAXINT - 1),
        "PositiveSmallIntegerField": (0, MAXINT - 1),
        "PositiveIntegerField": (0, MAXINT - 1),
        "SmallAutoField": (1, MAXINT - 1),
        "AutoField": (1, MAXINT - 1),
        "BigAutoField": (1, MAXINT - 1),
    }

    def bulk_batch_size(self, field, objs):
        # This value is used in cascade deletions, and also on bulk insertions
        # This is the limit of the number of entities that can be manipulated in
        # a single transaction

        settings_dict = self.connection.settings_dict
        if _BULK_BATCH_SIZE_SETTING in settings_dict['OPTIONS']:
            return int(settings_dict['OPTIONS'][_BULK_BATCH_SIZE_SETTING])
        return 500

    def quote_name(self, name):
        return name

    def date_trunc_sql(self, lookup_type, field_name, tzname=None):
        return ""

    def datetime_trunc_sql(self, lookup_type, field_name, tzname):
        return "", []

    def datetime_extract_sql(self, lookup_name, sql, tzname):
        return "", []

    def date_extract_sql(self, lookup_name, sql):
        return "", []

    def get_db_converters(self, expression):
        converters = super().get_db_converters(expression)

        db_type = expression.field.db_type(self.connection)
        internal_type = expression.field.get_internal_type()

        if internal_type == "TextField":
            converters.append(self.convert_textfield_value)
        elif internal_type == "DateTimeField":
            converters.append(self.convert_datetime_value)
        elif internal_type == "DateField":
            converters.append(self.convert_date_value)
        elif internal_type == "TimeField":
            converters.append(self.convert_time_value)
        elif internal_type == "DecimalField":
            converters.append(self.convert_decimal_value)
        elif internal_type == 'UUIDField':
            converters.append(self.convert_uuidfield_value)
        elif db_type == "list":
            converters.append(self.convert_list_value)
        elif db_type == "set":
            converters.append(self.convert_set_value)

        return converters

    def convert_uuidfield_value(self, value, expression, connection):
        if value is not None:
            value = uuid.UUID(value)
        return value

    def convert_textfield_value(self, value, expression, connection):
        if isinstance(value, bytes):
            # Should we log a warning here? It shouldn't have been stored as bytes
            value = value.decode("utf-8")
        return value

    def convert_datetime_value(self, value, expression, connection):
        return self.connection.ops.value_from_db_datetime(value)

    def convert_date_value(self, value, expression, connection):
        return self.connection.ops.value_from_db_date(value)

    def convert_time_value(self, value, expression, connection):
        return self.connection.ops.value_from_db_time(value)

    def convert_decimal_value(self, value, expression, connection):
        return self.connection.ops.value_from_db_decimal(value)

    def convert_list_value(self, value, expression, connection):
        if expression.output_field.db_type(connection) != "list":
            return value

        if not value:
            value = []
        return value

    def convert_set_value(self, value, expression, connection):
        if expression.output_field.db_type(connection) != "set":
            return value

        if not value:
            value = set()
        else:
            value = set(value)
        return value

    def sql_flush(self, style, tables, allow_cascade=False, reset_sequences=False, *args, **kwargs):
        additional_djangaeidx_tables = [
            x
            for x in self.connection.introspection.table_names()
            if [y for y in tables if x.startswith("_djangae_idx_{}".format(y))]
        ]

        return [
            FlushCommand(table, self.connection)
            for table in tables + additional_djangaeidx_tables
        ]

    def prep_lookup_key(self, model, value, field):
        if isinstance(value, str):
            value = value[:500]
            left = value[500:]
            if left:
                warnings.warn(
                    "Truncating primary key that is over 500 characters. " "THIS IS AN ERROR IN YOUR PROGRAM.",
                    RuntimeWarning,
                )

            # This is a bit of a hack. Basically when you query an integer PK with a
            # string containing an int. SQL seems to return the row regardless of type, and as far as
            # I can tell, Django at no point tries to cast the value to an integer. So, in the
            # case where the internal type is an AutoField, we try to cast the string value
            # I would love a more generic solution... patches welcome!
            # It would be nice to see the SQL output of the lookup_int_as_str test is on SQL, if
            # the string is converted to an int, I'd love to know where!
            if field.get_internal_type() == "AutoField":
                try:
                    value = int(value)
                except (TypeError, ValueError):
                    pass

            value = get_datastore_key(model, value)
        else:
            value = get_datastore_key(model, value)

        return value

    def prep_lookup_date(self, model, value, field):
        if isinstance(value, datetime.datetime):
            return value

        return self.adapt_datefield_value(value)

    def prep_lookup_time(self, model, value, field):
        if isinstance(value, datetime.datetime):
            return value

        return self.adapt_timefield_value(value)

    def prep_lookup_value(self, model, value, field, column=None):
        if field.primary_key and (not column or column == model._meta.pk.column):
            return self.prep_lookup_key(model, value, field)

        db_type = field.db_type(self.connection)

        if db_type == "date":
            return self.prep_lookup_date(model, value, field)
        elif db_type == "time":
            return self.prep_lookup_time(model, value, field)
        elif db_type in ("list", "set"):
            if hasattr(value, "__len__") and not value:
                value = None  # Convert empty lists to None
            elif hasattr(value, "__iter__"):
                # Convert sets to lists
                value = list(value)

        return value

    def value_for_db(self, value, field):
        if value is None:
            return None

        db_type = field.db_type(self.connection)

        if db_type in ("integer", "long"):
            if isinstance(value, float):
                # round() always returns a float, which has a smaller max value than an int
                # so only round() it if it's already a float
                value = round(value)

            value = int(value)
        elif db_type == "float":
            value = float(value)
        elif db_type == "string" or db_type == "text":
            value = coerce_unicode(value)
        elif db_type == "bytes":
            # Store BlobField, DictField and EmbeddedModelField values as Blobs.
            # We encode to bytes, as that's what the Cloud Datastore API expects
            # we use ASCII to make sure there's no funky unicode data, it should
            # be binary
            if isinstance(value, str):
                value = value.encode("ascii")
        elif db_type == "decimal":
            value = self.adapt_decimalfield_value(value, field.max_digits, field.decimal_places)
        elif db_type in ("list", "set"):
            if hasattr(value, "__len__") and not value:
                value = None  # Convert empty lists to None
            elif hasattr(value, "__iter__"):
                # Convert sets to lists
                value = list(value)

        return value

    def last_insert_id(self, cursor, db_table, column):
        return cursor.lastrowid

    def last_executed_query(self, cursor, sql, params):
        """
            We shouldn't have to override this, but Django's BaseOperations.last_executed_query
            assumes does u"QUERY = %r" % (sql) which blows up if you have encode unicode characters
            in your SQL. Technically this is a bug in Django for assuming that sql is ASCII but
            it's only our backend that will ever trigger the problem
        """
        return u"QUERY = {}".format(smart_text(sql))

    def fetch_returned_insert_id(self, cursor):
        return cursor.lastrowid

    def adapt_datetimefield_value(self, value):
        value = make_utc_compatible(value)
        return value

    def value_to_db_datetime(self, value):  # Django 1.8 compatibility
        return self.adapt_datetimefield_value(value)

    def adapt_datefield_value(self, value):
        if value is not None:
            value = datetime.datetime.combine(value, datetime.time())
        return value

    def adapt_timefield_value(self, value):
        if value is not None:
            value = make_utc_compatible(value)
            value = datetime.datetime.combine(datetime.datetime.utcfromtimestamp(0), value)
        return value

    def adapt_decimalfield_value(self, value, max_digits, decimal_places):
        if isinstance(value, decimal.Decimal):
            return decimal_to_string(value, max_digits, decimal_places)
        return value

    def value_to_db_decimal(self, value, max_digits, decimal_places):  # Django 1.8 compatibility
        return self.adapt_decimalfield_value(value, max_digits, decimal_places)

    # Unlike value_to_db, these are not overridden or standard Django, it's just nice to have symmetry
    def value_from_db_datetime(self, value):
        if isinstance(value, int):
            # App Engine Query's don't return datetime fields (unlike Get) I HAVE NO IDEA WHY
            value = ensure_datetime(value)

        if value is not None and settings.USE_TZ and timezone.is_naive(value):
            value = value.replace(tzinfo=timezone.utc)
        return value

    def value_from_db_date(self, value):
        if isinstance(value, int):
            # App Engine Query's don't return datetime fields (unlike Get) I HAVE NO IDEA WHY
            value = ensure_datetime(value)

        if value:
            value = value.date()
        return value

    def value_from_db_time(self, value):
        if isinstance(value, int):
            # App Engine Query's don't return datetime fields (unlike Get) I HAVE NO IDEA WHY
            value = ensure_datetime(value).time()

        if value is not None and settings.USE_TZ and timezone.is_naive(value):
            value = value.replace(tzinfo=timezone.utc)

        if value:
            value = value.time()

        return value

    def value_from_db_decimal(self, value):
        if value:
            value = decimal.Decimal(value)
        return value


class DatabaseClient(BaseDatabaseClient):
    pass


class DatabaseCreation(BaseDatabaseCreation):
    data_types = {
        "AutoField": "key",
        "RelatedAutoField": "key",
        "ForeignKey": "key",
        "OneToOneField": "key",
        "ManyToManyField": "key",
        "BigIntegerField": "long",
        "BooleanField": "bool",
        "CharField": "string",
        "CommaSeparatedIntegerField": "string",
        "DateField": "date",
        "DateTimeField": "datetime",
        "DecimalField": "decimal",
        "DurationField": "long",
        "EmailField": "string",
        "FileField": "string",
        "FilePathField": "string",
        "FloatField": "float",
        "ImageField": "string",
        "IntegerField": "integer",
        "IPAddressField": "string",
        "NullBooleanField": "bool",
        "PositiveIntegerField": "integer",
        "PositiveSmallIntegerField": "integer",
        "SlugField": "string",
        "SmallIntegerField": "integer",
        "TimeField": "time",
        "URLField": "string",
        "TextField": "text",
        "BinaryField": "bytes",
    }

    def __init__(self, *args, **kwargs):
        self.testbed = None
        super(DatabaseCreation, self).__init__(*args, **kwargs)

    def sql_create_model(self, model, *args, **kwargs):
        return [], {}

    def sql_for_pending_references(self, model, *args, **kwargs):
        return []

    def sql_indexes_for_model(self, model, *args, **kwargs):
        return []

    def _create_test_db(self, verbosity, autoclobber, *args):
        pass

    def _destroy_test_db(self, name, verbosity):
        pass


class DatabaseIntrospection(BaseDatabaseIntrospection):
    def get_table_list(self, cursor):
        query = cursor.connection.gclient.query(kind="__kind__")
        query.keys_only()
        kinds = [entity.key.id_or_name for entity in query.fetch()]
        return [TableInfo(x, "t") for x in kinds]

    def get_sequences(self, cursor, table_name, table_fields=()):
        # __key__ is the only column that can auto-populate
        return [{'table': table_name, 'column': '__key__'}]


class DatabaseSchemaEditor(BaseDatabaseSchemaEditor):
    def column_sql(self, model, field, include_default=False):
        return "", {}

    def create_model(self, model):
        """ Don't do anything when creating tables """
        pass

    def alter_unique_together(self, *args, **kwargs):
        pass

    def alter_field(self, from_model, from_field, to_field):
        pass

    def remove_field(self, from_model, field):
        pass

    def add_field(self, model, field):
        pass

    def alter_index_together(self, model, old_index_together, new_index_together):
        pass

    def delete_model(self, model):
        pass


class DatabaseFeatures(BaseDatabaseFeatures):
    empty_fetchmany_value = []
    supports_transactions = False  # FIXME: Make this True!
    can_return_id_from_insert = True
    supports_select_related = False
    autocommits_when_autocommit_is_off = True
    uses_savepoints = False
    allows_auto_pk_0 = False
    has_native_duration_field = False


class DatabaseWrapper(BaseDatabaseWrapper):

    data_types = DatabaseCreation.data_types  # These moved in 1.8

    operators = {
        "exact": "= %s",
        "iexact": "iexact %s",
        "contains": "contains %s",
        "icontains": "icontains %s",
        "regex": "regex %s",
        "iregex": "iregex %s",
        "gt": "> %s",
        "gte": ">= %s",
        "lt": "< %s",
        "lte": "<= %s",
        "startswith": "startswith %s",
        "endswith": "endswith %s",
        "istartswith": "istartswith %s",
        "iendswith": "iendswith %s",
        "isnull": "isnull %s",
    }

    Database = Database

    # These attributes are only used by Django >= 1.11
    client_class = DatabaseClient
    features_class = DatabaseFeatures
    introspection_class = DatabaseIntrospection
    features_class = DatabaseFeatures
    ops_class = DatabaseOperations
    creation_class = DatabaseCreation
    validation_class = BaseDatabaseValidation

    def __init__(self, *args, **kwargs):
        super(DatabaseWrapper, self).__init__(*args, **kwargs)

        if not hasattr(self, "client"):
            # Django 1.11 creates these automatically, when we call super
            # These are here for Django <= 1.10
            self.features = DatabaseFeatures(self)
            self.ops = DatabaseOperations(self)
            self.client = DatabaseClient(self)
            self.creation = DatabaseCreation(self)
            self.introspection = DatabaseIntrospection(self)
            self.validation = BaseDatabaseValidation(self)

        self.gcloud_project = self.settings_dict["PROJECT"]
        self.namespace = self.settings_dict.get("NAMESPACE") or None
        self.autocommit = True

    def is_usable(self):
        return True

    def get_connection_params(self):
        if not self.settings_dict.get("INDEXES_FILE"):
            raise ImproperlyConfigured("You must specify an INDEXES_FILE in the DATABASES setting")

        return self.settings_dict.copy()

    def get_new_connection(self, params):
        conn = Connection(self, params)
        load_special_indexes(conn)  # make sure special indexes are loaded
        return conn

    def init_connection_state(self):
        pass

    def _start_transaction_under_autocommit(self):
        pass

    def _set_autocommit(self, enabled):
        self.autocommit = enabled

    def create_cursor(self, name=None):
        self.name = name
        if not self.connection:
            self.connection = self.get_new_connection(self.settings_dict)

        return Cursor(self.connection)

    def schema_editor(self, *args, **kwargs):
        return DatabaseSchemaEditor(self, *args, **kwargs)

    def validate_no_broken_transaction(self):
        # Override this to do nothing, because it's not relevant to the Datastore
        pass

    def on_commit(self, func):
        # FIXME: This override needs to go away when we implement in_atomic_block
        # on the connection - which we can't do unless we move wholesale to Django's
        # atomic decorator (see issue #10)

        from django.db.transaction import TransactionManagementError
        from gcloudc.db.transaction import in_atomic_block

        if in_atomic_block(using=self.alias or "default"):
            # Transaction in progress; save for execution on commit.
            self.run_on_commit.append((set(self.savepoint_ids), func))
        elif not self.get_autocommit():
            raise TransactionManagementError('on_commit() cannot be used in manual transaction management')
        else:
            # No transaction in progress and in autocommit mode; execute
            # immediately.
            func()

    def run_and_clear_commit_hooks(self):
        # FIXME: This override needs to go away when we implement in_atomic_block
        # on the connection - which we can't do unless we move wholesale to Django's
        # atomic decorator (see issue #10)

        from gcloudc.db.transaction import in_atomic_block

        if in_atomic_block(using=self.alias or "default"):
            # FIXME: This should throw an error, but we can't because
            # Django's atomic blocks toggle autocommit in get_or_create
            # We need to unify with Django's decorators (see issue #10)
            return

        current_run_on_commit = self.run_on_commit
        self.run_on_commit = []
        while current_run_on_commit:
            sids, func = current_run_on_commit.pop(0)
            func()

    def close(self):
        """Close the connection to the database."""

        from gcloudc.db.transaction import in_atomic_block

        # FIXME: This override needs to go away when we implement in_atomic_block
        # on the connection - which we can't do unless we move wholesale to Django's
        # atomic decorator (see issue #10)

        self.validate_thread_sharing()
        self.run_on_commit = []

        # Don't call validate_no_atomic_block() to avoid making it difficult
        # to get rid of a connection in an invalid state. The next connect()
        # will reset the transaction state anyway.
        if self.closed_in_transaction or self.connection is None:
            return
        try:
            self._close()
        finally:
            if in_atomic_block(using=self.alias or "default"):
                self.closed_in_transaction = True
                self.needs_rollback = True
            else:
                self.connection = None



import datetime

from . import TestCase
from django import forms
from django.core import serializers
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.db import models
from gcloudc.db.models.fields.iterable import ListField, SetField
from gcloudc.db.models.fields.related import RelatedListField, RelatedSetField
from gcloudc.db.models.fields.charfields import CharField


class IterableIterableRelatedModel(models.Model):
    name = models.CharField(max_length=500)

    def __str__(self):
        return "%s:%s" % (self.pk, self.name)

    class Meta:
        app_label = "gcloudc"


class IterableIterableFieldsWithValidatorsModel(models.Model):
    set_field = SetField(models.CharField(max_length=100), min_length=2, max_length=3, blank=False)

    list_field = ListField(models.CharField(max_length=100), min_length=2, max_length=3, blank=False)

    related_set = RelatedSetField(IterableIterableRelatedModel, min_length=2, max_length=3, blank=False)

    related_list = RelatedListField(
        IterableIterableRelatedModel, related_name="iterable_list", min_length=2, max_length=3, blank=False
    )


class IterableFieldModel(models.Model):
    set_field = SetField(models.CharField(max_length=1))
    list_field = ListField(models.CharField(max_length=1))
    set_field_int = SetField(models.BigIntegerField(max_length=1))
    list_field_int = ListField(models.BigIntegerField(max_length=1))
    set_field_dt = SetField(models.DateTimeField())
    list_field_dt = ListField(models.DateTimeField())

    class Meta:
        app_label = "gcloudc"


class IterableFieldTests(TestCase):
    def test_filtering_on_iterable_fields(self):
        list1 = IterableFieldModel.objects.create(
            list_field=["A", "B", "C", "D", "E", "F", "G"], set_field=set(["A", "B", "C", "D", "E", "F", "G"])
        )
        list2 = IterableFieldModel.objects.create(
            list_field=["A", "B", "C", "H", "I", "J"], set_field=set(["A", "B", "C", "H", "I", "J"])
        )

        # filtering using __contains lookup with ListField:
        qry = IterableFieldModel.objects.filter(list_field__contains="A")
        self.assertEqual(sorted(x.pk for x in qry), sorted([list1.pk, list2.pk]))
        qry = IterableFieldModel.objects.filter(list_field__contains="H")
        self.assertEqual(sorted(x.pk for x in qry), [list2.pk])

        # filtering using __contains lookup with SetField:
        qry = IterableFieldModel.objects.filter(set_field__contains="A")
        self.assertEqual(sorted(x.pk for x in qry), sorted([list1.pk, list2.pk]))
        qry = IterableFieldModel.objects.filter(set_field__contains="H")
        self.assertEqual(sorted(x.pk for x in qry), [list2.pk])

        # filtering using __overlap lookup with ListField:
        qry = IterableFieldModel.objects.filter(list_field__overlap=["A", "B", "C"])
        self.assertEqual(sorted(x.pk for x in qry), sorted([list1.pk, list2.pk]))
        qry = IterableFieldModel.objects.filter(list_field__overlap=["H", "I", "J"])
        self.assertEqual(sorted(x.pk for x in qry), sorted([list2.pk]))

        # filtering using __overlap lookup with SetField:
        qry = IterableFieldModel.objects.filter(set_field__overlap=set(["A", "B"]))
        self.assertEqual(sorted(x.pk for x in qry), sorted([list1.pk, list2.pk]))
        qry = IterableFieldModel.objects.filter(set_field__overlap=["H"])
        self.assertEqual(sorted(x.pk for x in qry), [list2.pk])

    def test_empty_iterable_fields(self):
        """ Test that an empty set field always returns set(), not None """
        instance = IterableFieldModel()
        # When assigning
        self.assertEqual(instance.set_field, set())
        self.assertEqual(instance.list_field, [])
        instance.save()

        instance = IterableFieldModel.objects.get()
        # When getting it from the db
        self.assertEqual(instance.set_field, set())
        self.assertEqual(instance.list_field, [])

    def test_list_field(self):
        instance = IterableFieldModel.objects.create()
        self.assertEqual([], instance.list_field)
        instance.list_field.append("One")
        self.assertEqual(["One"], instance.list_field)
        instance.save()

        self.assertEqual(["One"], instance.list_field)

        instance = IterableFieldModel.objects.get(pk=instance.pk)
        self.assertEqual(["One"], instance.list_field)

        results = IterableFieldModel.objects.filter(list_field__contains="One")
        self.assertEqual([instance], list(results))

        self.assertEqual([1, 2], ListField(models.IntegerField).to_python("[1, 2]"))

    def test_set_field(self):
        instance = IterableFieldModel.objects.create()
        self.assertEqual(set(), instance.set_field)
        instance.set_field.add("One")
        self.assertEqual(set(["One"]), instance.set_field)
        instance.save()

        self.assertEqual(set(["One"]), instance.set_field)

        instance = IterableFieldModel.objects.get(pk=instance.pk)
        self.assertEqual(set(["One"]), instance.set_field)

        self.assertEqual({1, 2}, SetField(models.IntegerField).to_python("[1, 2]"))

    def test_empty_list_queryable_with_is_null(self):
        instance = IterableFieldModel.objects.create()

        self.assertTrue(IterableFieldModel.objects.filter(set_field__isempty=True).exists())

        instance.set_field.add(1)
        instance.save()

        self.assertFalse(IterableFieldModel.objects.filter(set_field__isempty=True).exists())
        self.assertTrue(IterableFieldModel.objects.filter(set_field__isempty=False).exists())

        self.assertFalse(IterableFieldModel.objects.exclude(set_field__isempty=False).exists())
        self.assertTrue(IterableFieldModel.objects.exclude(set_field__isempty=True).exists())

    def test_serialization(self):
        dt = datetime.datetime(2017, 1, 1, 12)
        instance = IterableFieldModel.objects.create(
            set_field={u"foo"},
            list_field=[u"bar"],
            set_field_int={123},
            list_field_int=[456],
            set_field_dt={dt},
            list_field_dt=[dt],
        )

        self.assertEqual("['foo']", instance._meta.get_field("set_field").value_to_string(instance))
        self.assertEqual("['bar']", instance._meta.get_field("list_field").value_to_string(instance))
        self.assertEqual("[123]", instance._meta.get_field("set_field_int").value_to_string(instance))
        self.assertEqual("[456]", instance._meta.get_field("list_field_int").value_to_string(instance))
        self.assertEqual("['2017-01-01T12:00:00']", instance._meta.get_field("set_field_dt").value_to_string(instance))
        self.assertEqual("['2017-01-01T12:00:00']", instance._meta.get_field("list_field_dt").value_to_string(instance))

    def test_saving_forms(self):
        class TestForm(forms.ModelForm):
            class Meta:
                model = IterableFieldModel
                fields = ("set_field", "list_field")

        post_data = {"set_field": ["1", "2"], "list_field": ["1", "2"]}

        form = TestForm(post_data)
        self.assertTrue(form.is_valid())
        self.assertTrue(form.save())

    def test_cannot_have_min_length_and_blank(self):
        """ Having min_length=X, blank=True doesn't make any sense, especially when you consider
            that django will skip the min_length check when the value (list/set)is empty.
        """
        self.assertRaises(ImproperlyConfigured, ListField, CharField(max_length=100), min_length=1, blank=True)
        self.assertRaises(ImproperlyConfigured, SetField, CharField(max_length=100), min_length=1, blank=True)

    def test_list_field_set_field_min_max_lengths_valid(self):
        """ Test that when the min_legnth and max_length of a ListField and SetField are correct
            that no validation error is rasied.
        """
        others = []
        for x in range(2):
            others.append(IterableIterableRelatedModel.objects.create())
        instance = IterableIterableFieldsWithValidatorsModel(
            related_set=set(others),  # not being tested here
            related_list=others,  # not being tested here
            set_field=set(["1", "2"]),
            list_field=["1", "2"],
        )
        instance.full_clean()

    def test_list_field_max_length_invalid(self):
        others = []
        for x in range(2):
            others.append(IterableIterableRelatedModel.objects.create())
        instance = IterableIterableFieldsWithValidatorsModel(
            related_set=set(others),  # not being tested here
            related_list=others,  # not being tested here
            set_field=set(["1", "2"]),  # not being tested here
            list_field=["1", "2", "3", "4", "5"],
        )
        self.assertRaisesMessage(
            ValidationError,
            "{'list_field': ['Ensure this field has at most 3 items (it has 5).']}",
            instance.full_clean,
        )

    def test_list_field_min_length_invalid(self):
        others = []
        for x in range(2):
            others.append(IterableIterableRelatedModel.objects.create())
        instance = IterableIterableFieldsWithValidatorsModel(
            related_set=set(others),  # not being tested here
            related_list=others,  # not being tested here
            set_field=set(["1", "2"]),  # not being tested here
            list_field=["1"],
        )
        self.assertRaisesMessage(
            ValidationError,
            "{'list_field': ['Ensure this field has at least 2 items (it has 1).']}",
            instance.full_clean,
        )

    def test_set_field_max_length_invalid(self):
        others = []
        for x in range(2):
            others.append(IterableIterableRelatedModel.objects.create())
        instance = IterableIterableFieldsWithValidatorsModel(
            related_set=set(others),  # not being tested here
            related_list=others,  # not being tested here
            list_field=["1", "2"],  # not being tested here
            set_field=set(["1", "2", "3", "4", "5"]),
        )
        self.assertRaisesMessage(
            ValidationError, "{'set_field': ['Ensure this field has at most 3 items (it has 5).']}", instance.full_clean
        )

    def test_set_field_min_length_invalid(self):
        others = []
        for x in range(2):
            others.append(IterableIterableRelatedModel.objects.create())
        instance = IterableIterableFieldsWithValidatorsModel(
            related_set=set(others),  # not being tested here
            related_list=others,  # not being tested here
            list_field=["1", "2"],  # not being tested here
            set_field=set(["1"]),
        )
        self.assertRaisesMessage(
            ValidationError,
            "{'set_field': ['Ensure this field has at least 2 items (it has 1).']}",
            instance.full_clean,
        )

    def test_list_field_serializes_and_deserializes(self):
        obj = IterableFieldModel(list_field=["foo", "bar"])
        data = serializers.serialize("json", [obj])

        new_obj = next(serializers.deserialize("json", data)).object
        self.assertEqual(new_obj.list_field, ["foo", "bar"])

    def test_set_field_serializes_and_deserializes(self):
        obj = IterableFieldModel(set_field=set(["foo", "bar"]))
        data = serializers.serialize("json", [obj])

        new_obj = next(serializers.deserialize("json", data)).object
        self.assertEqual(new_obj.set_field, set(["foo", "bar"]))
